# -*- coding: utf-8 -*-
# @Time    : 04/09/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : clustering_algorithms.py
# @Software: PyCharm

import warnings
import pandas as pd
import inspect

warnings.simplefilter(action='ignore', category=FutureWarning)
pd.options.mode.chained_assignment = None  # default='warn'

import pandas as pd
import numpy as np
from sklearn.preprocessing import MinMaxScaler
from sklearn.cluster import KMeans, MiniBatchKMeans, MeanShift, DBSCAN, Birch, OPTICS, AffinityPropagation, \
    AgglomerativeClustering, SpectralClustering
from yellowbrick.cluster import KElbowVisualizer
from openpy_fxts.models.anomaly.removal_data import seasonal
import math
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import warnings
import os

from tslearn.clustering import TimeSeriesKMeans
from tslearn.barycenters import euclidean_barycenter, dtw_barycenter_averaging, dtw_barycenter_averaging_subgradient, \
    softdtw_barycenter

from sklearn.preprocessing import StandardScaler

from tslearn.piecewise import PiecewiseAggregateApproximation
from tslearn.piecewise import SymbolicAggregateApproximation
from tslearn.piecewise import OneD_SymbolicAggregateApproximation

from sklearn.metrics import silhouette_score, davies_bouldin_score, calinski_harabasz_score
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score, mean_absolute_percentage_error
from sklearn.manifold import TSNE, MDS
from sklearn.cluster import KMeans, MiniBatchKMeans, DBSCAN, Birch, OPTICS, AffinityPropagation, \
    AgglomerativeClustering, SpectralClustering
from sklearn.cluster import MeanShift, estimate_bandwidth
from sklearn.mixture import GaussianMixture
from minisom import MiniSom

from scipy import stats

from scipy.cluster.hierarchy import dendrogram
from tqdm.autonotebook import tqdm
from .data_preprocessing import moving_window_mean

from sklearn.neighbors import NearestNeighbors

clu_no_centroid = ['DBSCAN', 'Birch', 'GaussianMixture', 'OPTICS', 'Agglomerative', 'Spectral']
model_with_eps = ['DBSCAN', 'OPTICS']


def set_tittle_cluster(key, gamma):
    if key == 'centroid':
        return 'Centroid'
    if key == 'mean':
        return 'Mean'
    if key == 'euclidean':
        return "Euclidean"
    if key == 'dba_vectorized':
        return "DBA (vectorized version of Petitjean's EM)"
    if key == 'dba_subgradient':
        return "DBA (subgradient descent approach)"
    if key == 'soft_dtw':
        return f"Soft-DTW barycenter ($\gamma$={gamma})"


def set_tittle_cluster(case: bool = None, gamma=None):
    if case == 'centroid':
        return 'Centroid'
    if case == 'mean':
        return 'Mean'
    if case == 'euclidean':
        return "Euclidean"
    if case == 'dba_vectorized':
        return "DBA (vectorized \nversion of Petitjean's EM)"
    if case == 'dba_subgradient':
        return "DBA (subgradient \ndescent approach)"
    if case == 'soft_dtw':
        return f"Soft-DTW barycenter\n ($\gamma$={gamma})"


class clustering_ts:

    def __init__(
            self,
            users=None,
            model: str = 'KMeans',
            metric: str = 'dtw',
            type_dr: str = None,
            cdata: str = None,
            reduced: bool = False,
            n_components: int = 2,
            seed: int = 23,
            df_sce: pd.DataFrame = None,
            dict_sce: dict = None,
            clu_lbl=None,
            n_jobs: int = -1,
            max_iter: int = 10,
            is_hours: bool = None,
            is_month: bool = None
    ):

        if users is None:
            users = df_sce.columns[:-11]
        self.model = model
        self.users = users
        self.metric = metric
        self.type_dr = type_dr
        self.cdata = cdata
        self.reduced = reduced
        self.n_components = n_components
        self.seed = seed
        self.df_sce = df_sce
        self.dict_sce = dict_sce
        self.clu_lbl = clu_lbl
        self.n_jobs = -1
        self.max_iter = 10
        self.is_hours = is_hours
        self.is_month = is_month
        self.acum_bar = None

    def reshape_to_clu_atypical_mean(
            self,
            plot_user: bool = False,
            id_user: str = None,
            op: int = 1
    ):
        dict_sce = dict()
        df_users_mean = pd.DataFrame()
        if self.reduced is False and self.clu_lbl is None:
            for k in self.users:
                df_aux = self.get_df_aux(k)
                dict_sce[k] = df_aux
            return dict_sce
        if self.reduced and self.clu_lbl is None:
            for k in self.users:
                df_aux = self.get_df_aux(k)
                dict_sce[k] = df_aux
                user_mean = self.outlier_removal(df_aux, df_aux.columns, op, k)
                df_users_mean = pd.concat([df_users_mean, user_mean], axis=0)
            df_users_mean.interpolate(inplace=True, axis=1)
            dict_sce['means'] = df_users_mean
            return dict_sce
        if self.reduced and self.clu_lbl is not None:
            for k in self.users:
                df_aux = self.get_df_aux(k)
                cols = df_aux.columns
                df_aux = pd.merge(df_aux, self.clu_lbl, left_index=True, right_index=True)
                df_aux.rename(columns={"cluster": "clu_1"}, inplace=True)
                dict_sce[k] = df_aux
                for x in self.clu_lbl.unique():
                    df_clu_aux = self.outlier_removal(df_aux[df_aux.clu_1 == x], cols, op, k)
                    df_clu_aux.loc[:, 'clu_1'] = x
                    df_users_mean = pd.concat([df_users_mean, df_clu_aux], axis=0)
            dict_sce['means'] = df_users_mean
            return dict_sce
        if self.reduced is False and self.clu_lbl is not None:
            for k in self.users:
                df_aux = self.get_df_aux(k)
                df_aux = pd.merge(df_aux, self.clu_lbl, left_index=True, right_index=True)
                df_aux.rename(columns={"cluster": "clu_1"}, inplace=True)
                dict_sce[k] = df_aux
            return dict_sce

        df_users_mean_1, df_users_mean_2 = pd.DataFrame(), pd.DataFrame()
        for k in self.users:
            df_aux = self.get_df_aux(k)
            if self.clu_lbl is not None:
                df_aux = pd.merge(df_aux, self.clu_lbl, left_index=True, right_index=True)
                # df_aux.loc[:, 'user'] = k
                df_aux.rename(columns={"cluster": "clu_1"}, inplace=True)
            dict_sce[k] = df_aux
            if self.reduced and self.clu_lbl is not None:
                cols = df_aux.columns[:-1]
                for x in self.clu_lbl.unique():
                    df_clu_aux = self.outlier_removal(df_aux[df_aux.clu_1 == x], cols, op, k)
                    df_clu_aux.loc[:, 'clu_1'] = x
                    df_users_red_2 = pd.concat([df_users_mean_2, df_clu_aux], axis=0)
            else:
                if self.clu_lbl is not None:
                    cols = df_aux.columns[:-1]
                else:
                    cols = df_aux.columns
                user_mean = self.outlier_removal(df_aux, cols, op, k)
                df_users_mean_1 = pd.concat([df_users_mean_1, user_mean], axis=0)
        df_users_mean_1.interpolate(inplace=True, axis=1)
        dict_sce['means'] = df_users_mean_1

        if self.reduced and self.clu_lbl is not None:
            dict_sce['means'] = df_users_mean_2

        if plot_user:
            if id_user is None:
                cols = 3
                rows = round((len(dict_sce.keys()) - 1) / cols) + 1
                fig_1, ax = plt.subplots(nrows=rows, ncols=cols, sharex=True)
                n, m = 0, 0
                for key, value in dict_sce.items():
                    if key == 'means':
                        pass
                    else:
                        ax[n, m].plot(dict_sce[key].T, "k-", alpha=0.2)
                        ax[n, m].plot(dict_sce['means'].loc[key].T, c='r')
                        ax[n, m].set_title(key)
                        m += 1
                        if m == cols:
                            m = 0
                            n += 1
                plt.show()
            else:
                plt.plot(dict_sce[id_user].T, "k-", alpha=0.2)
                plt.plot(dict_sce['means'].loc[id_user].T, c='r')
                plt.title(id_user)
                plt.show()

        return dict_sce

    def get_df_aux(self, user):
        data = self.df_sce.loc[:, [user, 'count_day', 'date']].copy()
        aux = []
        for i in data['count_day'].unique():
            aux.append(data[data['count_day'] == i].shape[0])
        n_meas = max(aux)
        data = data.reset_index()
        if self.is_hours:
            data[data.columns[0]] = data[data.columns[0]].apply(lambda x: x.strftime('%H:%M'))
        elif self.is_month:
            data[data.columns[0]].apply(lambda x: x.strftime('%m-%Y'))
        else:
            data[data.columns[0]] = data[data.columns[0]].apply(lambda x: x.strftime('%H:%M'))
        data.set_index(data.columns[0], inplace=True)
        day_aux = list(data.count_day.value_counts()[data.count_day.value_counts() == n_meas].index)
        data = data[data.count_day.isin(day_aux)]
        data.reset_index(inplace=True)
        data.drop(columns=['count_day'], inplace=True)
        if self.is_hours:
            df_aux = pd.pivot(
                data,
                index=['date'],
                columns=['index']
            )
            df_aux.columns = ['_'.join(col) for col in df_aux.columns.values]
            df_aux.columns = [item.replace(f'{user}_', "") for item in df_aux.columns]
            return df_aux
        elif self.is_month:
            df_aux = pd.DataFrame(data.T.loc[user]).T
            df_aux.columns = [str(i) for i in df_aux.columns.values]
            return df_aux

    def outlier_removal(self, df_aux, cols, op=1, user=None):
        data_graf = []
        vector_Prom = []
        aux = 0
        for i in cols:
            data = df_aux[i]
            size = len(data)  # determinar la longitud de los datos
            if size != 0:
                if op == 1:
                    # print(k, i)
                    Q1 = data.quantile(0.25)
                    Q3 = data.quantile(0.75)
                    IQR = Q3 - Q1
                    lower_lim = Q1 - 1.5 * IQR
                    upper_lim = Q3 + 1.5 * IQR
                    outliers_low = (data < lower_lim)
                    outliers_up = (data > upper_lim)
                    data[outliers_low] = lower_lim
                    data[outliers_up] = upper_lim
                if op == 2:
                    z_critical = stats.norm.ppf(q=0.999)
                    mean = data.mean()  # determina la media.
                    std = data.std()  # Get the population standard deviation
                    margin_of_error = z_critical * (std / math.sqrt(size))  # margen de error
                    confidence_interval = (mean - margin_of_error, mean + margin_of_error)
                    int_min = mean - margin_of_error
                    int_max = mean + margin_of_error
                    data = data[(data >= int_min)]  # filtrar obteniendo los valores de mayores al min
                    data = data[(data <= int_max)]  # filtrar obteniendo los valores de menores al min

            data_graf.append(data)
            df_mean = data_graf[aux].mean()
            vector_Prom.append(df_mean)
            aux += 1

        user_mean = pd.DataFrame(vector_Prom, index=cols, columns=[user]).T
        user_mean.interpolate(inplace=True, axis=1)
        return user_mean

    def transformer_dataset(
            self,
            model: str = None,  # 'PAA', 'SAX', '1d_SAX'
            n_paa_segments=5,
            n_sax_symbols=4,
            n_sax_symbols_avg=4,
            n_sax_symbols_slope=4
    ):

        prepo_ts = self.reshape_to_clu_atypical_mean()
        dict_scaler = self._scaler_data(prepo_ts)
        if dict_scaler['dataset'].isnull().sum().sum() > 0:
            dict_scaler['dataset'].ffill(inplace=True)

        # PAA transform (and inverse transform) of the data
        paa = PiecewiseAggregateApproximation(n_segments=n_paa_segments)
        paa_dataset_inv = paa.inverse_transform(paa.fit_transform(dict_scaler['dataset']))

        # SAX transform
        sax = SymbolicAggregateApproximation(n_segments=n_paa_segments, alphabet_size_avg=n_sax_symbols)
        sax_dataset_inv = sax.inverse_transform(sax.fit_transform(dict_scaler['dataset']))
        # 1d-SAX transform
        one_d_sax = OneD_SymbolicAggregateApproximation(
            n_segments=n_paa_segments,
            alphabet_size_avg=n_sax_symbols_avg,
            alphabet_size_slope=n_sax_symbols_slope)
        transformed_data = one_d_sax.fit_transform(dict_scaler['dataset'])
        one_d_sax_dataset_inv = one_d_sax.inverse_transform(transformed_data)

        i = 0
        plt.figure()
        plt.subplot(2, 2, 1)  # First, raw time series
        plt.plot(dict_scaler['dataset'].iloc[i].ravel(), "b-")
        plt.title("Raw time series")

        plt.subplot(2, 2, 2)  # Second, PAA
        plt.plot(dict_scaler['dataset'].iloc[i].ravel(), "b-", alpha=0.4)
        plt.plot(paa_dataset_inv[i].ravel(), "b-")
        plt.title("PAA")

        plt.subplot(2, 2, 3)  # Then SAX
        plt.plot(dict_scaler['dataset'].iloc[i].ravel(), "b-", alpha=0.4)
        plt.plot(sax_dataset_inv[i].ravel(), "b-")
        plt.title("SAX, %d symbols" % n_sax_symbols)

        plt.subplot(2, 2, 4)  # Finally, 1d-SAX
        plt.plot(dict_scaler['dataset'].iloc[i].ravel(), "b-", alpha=0.4)
        plt.plot(one_d_sax_dataset_inv[i].ravel(), "b-")
        plt.title(
            "1d-SAX, %d symbols""(%dx%d)" % (
                n_sax_symbols_avg * n_sax_symbols_slope, n_sax_symbols_avg, n_sax_symbols_slope
            )
        )
        plt.tight_layout()
        plt.show()
        # Scaler data
        dict_scaler = self._scaler_data(prepo_ts)
        if dict_scaler['dataset'].isnull().sum().sum() > 0:
            dict_scaler['dataset'].ffill(inplace=True)
        return dict_scaler

    def prepare_data_to_clu(
            self,
            atypical: bool = True
    ):
        data = self.df_sce.copy()
        aux = []
        for i in data['count_day'].unique():
            aux.append(data[data['count_day'] == i].shape[0])
        n_meas = max(aux)
        data = data.reset_index()
        data[data.columns[0]] = data[data.columns[0]].apply(lambda x: x.strftime('%H:%M'))
        data.set_index(data.columns[0], inplace=True)
        day_aux = list(data.count_day.value_counts()[data.count_day.value_counts() == n_meas].index)
        data = data[data.count_day.isin(day_aux)]
        if atypical:
            if type(self.users) == str:
                self.users = self.users.split()
            # remove atypical datasets
            for col in self.users:
                Q1 = data.loc[:, col].quantile(0.25)
                Q3 = data.loc[:, col].quantile(0.75)
                IQR = Q3 - Q1
                lower_lim = Q1 - 1.5 * IQR
                upper_lim = Q3 + 1.5 * IQR
                outliers_low = (data.loc[:, col] < lower_lim)
                outliers_up = (data.loc[:, col] > upper_lim)
                data.loc[:, col][(outliers_low | outliers_up)] = np.nan
                '''
                data.loc[:, col] = data.loc[:, col].apply(
                    lambda x: seasonal.nullify_outliers(x, period=n_meas),
                    axis='index'
                )
                '''
                if data[col].isna().sum().sum() > 0:
                    data[col].ffill(inplace=True)
        data.reset_index(inplace=True)
        # data['user'] = data.columns[1]
        sce_days = dict()
        for user in self.users:
            sce_days[user] = pd.pivot(
                data,
                values=user,
                index=['date'],
                columns=data.columns[0]
            )
        return sce_days

    def _scaler_data(self, dict_data, label=None):
        dict_scaler = {}
        if self.reduced and self.clu_lbl is None:
            df_scaler = dict_data['means']
            user_aux = list(dict_data['means'].index)
            cols = df_scaler.columns
        if self.reduced is False and self.clu_lbl is None:
            df_scaler = pd.DataFrame()
            user_aux = []
            for user, df_data in dict_data.items():
                if user != 'means':
                    if self.clu_lbl is not None:
                        cols = df_data.columns[:-1]
                        df_data = df_data[df_data.clu_1 == label].iloc[:, :len(cols)]
                    else:
                        cols = df_data.columns
                    df_scaler = pd.concat([df_scaler, df_data], axis=0)
                    user_aux += [user] * df_data.shape[0]
        if self.reduced and self.clu_lbl is not None:
            df_scaler = dict_data['means']
            cols = df_scaler.columns[:-1]
            df_scaler = df_scaler[df_scaler.clu_1 == label].iloc[:, :len(cols)]
            user_aux = list(df_scaler.index)
        if self.reduced is False and self.clu_lbl is not None:
            df_scaler = pd.DataFrame()
            user_aux = []
            for user, df_data in dict_data.items():
                if user != 'means':
                    if self.clu_lbl is not None:
                        cols = df_data.columns[:-1]
                        df_data = df_data[df_data.clu_1 == label].iloc[:, :len(cols)]
                    else:
                        cols = df_data.columns
                    df_scaler = pd.concat([df_scaler, df_data], axis=0)
                    user_aux += [user] * df_data.shape[0]
        scalars = {}
        for i in cols:
            scaler = MinMaxScaler(feature_range=(0, 1))
            s_s = scaler.fit_transform(df_scaler[i].values.reshape(-1, 1))
            s_s = np.reshape(s_s, len(s_s))
            scalars['scaler_' + i] = scaler
            df_scaler[i] = s_s
        dict_scaler['dataset'] = df_scaler
        dict_scaler['scaler'] = scalars
        dict_scaler['user'] = user_aux
        return dict_scaler

        if self.reduced:
            df_scaler = dict_data['means']
            user_aux = list(dict_data['means'].index)
        else:
            df_scaler = pd.DataFrame()
            user_aux = []
            for user, df_data in dict_data.items():
                if user != 'means':
                    if self.clu_lbl is not None:
                        cols = df_data.columns[:-1]
                        df_data = df_data[df_data.clu_1 == label].iloc[:, :len(cols)]
                    else:
                        cols = df_data.columns
                    df_scaler = pd.concat([df_scaler, df_data], axis=0)
                    user_aux += [user] * df_data.shape[0]
        scalars = {}
        for i in cols:
            scaler = MinMaxScaler(feature_range=(0, 1))
            s_s = scaler.fit_transform(df_scaler[i].values.reshape(-1, 1))
            s_s = np.reshape(s_s, len(s_s))
            scalars['scaler_' + i] = scaler
            df_scaler[i] = s_s
        dict_scaler['dataset'] = df_scaler
        dict_scaler['scaler'] = scalars
        dict_scaler['user'] = user_aux
        return dict_scaler

    def _inverse_transform(
            self,
            yhat,
            dict_scaler,
    ):
        for index, i in enumerate(dict_scaler['dataset'].columns):
            scaler = dict_scaler['scaler']['scaler_' + i]
            yhat[index, :] = scaler.inverse_transform(yhat[index, :].reshape(-1, 1)).reshape(yhat.shape[1], )
        return yhat

    def scaler_data(self, data):
        from sklearn.preprocessing import StandardScaler, MinMaxScaler
        scaler = MinMaxScaler()
        dict_scaler = {
            'dataset': scaler.fit_transform(data.T).T,
            'scaler': scaler
        }
        return dict_scaler

    def get_centroids_for_clustering(
            self,
            labels_clu,
            max_iter: int = 10,
            tol: float = 1e-3,
            gamma: float = 1.0,
            verbose: bool = False,
            all_barycenter: bool = False,
            mean: bool = None,
            euclidean: bool = None,
            dba_vec: bool = None,
            dba_sub: bool = None,
            soft_dtw: bool = None,
            all_plots: bool = False,
            plot_n_cluster: int = 0,
            plt_barycenters: bool = True,
            plt_bar_cluster: bool = None,
            plt_bar_summary: bool = None,
            plt_clusters: bool = None,
            view_clusters: bool = None,
            pct_clusters: bool = None
    ):

        if all_barycenter:
            mean, euclidean, dba_vec, dba_sub, soft_dtw = True, True, True, True, True
        if plt_barycenters:
            plt_bar_cluster, plt_bar_summary, = True, True
        if plt_clusters:
            view_clusters, pct_clusters = True, True
        dict_scaler = self.transformer_dataset()
        dict_scaler['dataset'].loc[:, 'cluster'] = labels_clu
        dataset = dict_scaler['dataset']
        ts_clustered = [0] * len(labels_clu.unique())
        for i in labels_clu.unique():
            ts_clustered[i] = np.array(dataset[dataset.cluster == i].iloc[:, :-1])
        dataset.drop(columns=['cluster'], inplace=True)
        dict_bar = {
            'dataset': dataset,
            'clusters': [x.T for x in ts_clustered]
        }
        # Find barycentre's
        bar = barycenter(ts_clustered, max_iter, tol, verbose, gamma)
        if mean:
            dict_bar['mean'] = bar.mean()
        if euclidean:
            dict_bar['euclidean'] = bar.euclidean()
        if dba_vec:
            dict_bar['dba_vectorized'] = bar.dba_vectorized()
        if dba_sub:
            dict_bar['dba_subgradient'] = bar.dba_subgradient()
        if soft_dtw:
            dict_bar['soft_dtw'] = bar.soft_dtw()

        for key, values_scales in dict_bar.items():
            if key == 'dataset':
                dict_bar['dataset'] = pd.DataFrame(
                    self._inverse_transform(values_scales.values.T, dict_scaler).T,
                    columns=dict_scaler['dataset'].columns,
                    index=dict_scaler['dataset'].index
                )
            else:
                for x in values_scales:
                    self._inverse_transform(x, dict_scaler)

        n_clusters = len(labels_clu.unique())

        if plt_bar_summary:
            fx_clu_plots.all_barycenter(True, False, plot_n_cluster, gamma, dict_bar)
        if plt_bar_cluster:
            fx_clu_plots.all_barycenter(False, True, plot_n_cluster, gamma, dict_bar)
        if view_clusters:
            self._view_cluster(n_clusters, dict_bar, gamma)
        if pct_clusters:
            fx_clu_plots.pct_clusters(labels_clu)
        dict_bar['dataset'].loc[:, 'cluster'] = labels_clu
        dict_bar['dataset']['user'] = dict_scaler['user']
        return dict_bar

    def train_clu_model(
            self,
            n_clusters: int = 3,
            n_components: int = 3,
            quantile: float = 0.13,
            n_samples: int = None,
            eps: float = 1.1,
            min_samples=None,
            threshold: float = 1.5,
            branching_factor: int = 50,
            damping: float = 0.90,
            sigma: float = 0.5,
            lr: float = 0.5,
            num_iteration: int = 500,
            points_2d: bool = False,
            view_clu: bool = False,
            acum_bar: bool = True
    ):
        """

        :param n_clusters:
        :param n_components:
        :param quantile:
        :param n_samples:
        :param eps:
        :param min_samples:
        :param threshold:
        :param branching_factor:
        :param damping:
        :param sigma:
        :param lr:
        :param num_iteration:
        :param points_2d:
        :param view_clu:
        :param acum_bar:
        :return:
        """

        parameters = {
            'type_dr': self.type_dr,
            'model': self.model,
            'metric': self.metric,
            'n_jobs': self.n_jobs,
            'max_iter': self.max_iter,
            'seed': self.seed,
            'n_clusters': n_clusters,
            'n_components': n_components,
            'quantile': quantile,
            'n_samples': n_samples,
            'eps': eps,
            'min_samples': min_samples,
            'threshold': threshold,
            'branching_factor': branching_factor,
            'damping': damping,
            'sigma': sigma,
            'lr': lr,
            'num_iteration': num_iteration,
            'points_2d': points_2d,
            'view_clu': view_clu
        }
        self.acum_bar = acum_bar
        mdl_clu = _choice_clu_model(parameters)
        if self.reduced is False and self.clu_lbl is None:
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_scaler = self._scaler_data(dict_sce)
            ts_scaler = dict_scaler['dataset'].values
            # Dimensional reduction
            if self.type_dr is not None:
                data = self.dimensionality_reduction().fit_transform(ts_scaler)
            else:
                data = ts_scaler
            return mdl_clu.check_model(data, ts_scaler, dict_scaler)
        if self.reduced and self.clu_lbl is None:
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_scaler = self._scaler_data(dict_sce)
            ts_scaler = dict_scaler['dataset'].values
            # Dimensional reduction
            if self.type_dr is not None:
                data = self.dimensionality_reduction().fit_transform(ts_scaler)
            else:
                data = ts_scaler
            return mdl_clu.check_model(data, ts_scaler, dict_scaler)
        if self.reduced and self.clu_lbl is not None:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
                models_with_labels[i] = mdl_clu.check_model(data, ts_scaler, dict_scaler)
            return models_with_labels
        if self.reduced is False and self.clu_lbl is not None:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
                models_with_labels[i] = mdl_clu.check_model(data, ts_scaler, dict_scaler)
            return models_with_labels

        if self.clu_lbl is not None and self.reduced:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
                models_with_labels[i] = mdl_clu.check_model(data, ts_scaler, dict_scaler)
            return models_with_labels
        if self.clu_lbl is not None:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
                models_with_labels[i] = mdl_clu.check_model(data, ts_scaler, dict_scaler)
            return models_with_labels
        else:
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_scaler = self._scaler_data(dict_sce)
            ts_scaler = dict_scaler['dataset'].values
            # Dimensional reduction
            if self.type_dr is not None:
                data = self.dimensionality_reduction().fit_transform(ts_scaler)
            else:
                data = ts_scaler
            return mdl_clu.check_model(data, ts_scaler, dict_scaler)

    def cluster_ts_extraction(
            self,
            n_clusters: int = 3,
            plot_n_cluster: int = 0,
            max_iter: int = 10,
            tol: float = 1e-3,
            gamma: float = 1.0,
            verbose: bool = False,
            all_barycenters: bool = False,
            centroid: bool = None,
            mean: bool = None,
            euclidean: bool = None,
            dba_vec: bool = None,
            dba_sub: bool = None,
            soft_dtw: bool = None,
            plt_barycenters: bool = None,
            plt_all_barycenters: bool = None,
            summary_barycenters: bool = None,
            plt_clu: bool = None,
            view_clusters: bool = None,
            pct_clusters: bool = None
    ):
        bool_bar = {
            'centroid': centroid,
            'mean': mean,
            'euclidean': euclidean,
            'dba_vec': dba_vec,
            'dba_sub': dba_sub,
            'soft_dtw': soft_dtw
        }
        bool_plots = {
            'plot_n_cluster': plot_n_cluster,
            'plt_barycenters': plt_barycenters,
            'plt_all_barycenters': plt_all_barycenters,
            'summary_barycenters': summary_barycenters,
            'plt_clu': plt_clu,
            'view_clusters': view_clusters,
            'pct_clusters': pct_clusters
        }
        if all_barycenters:
            bool_bar['centroid'] = True
            bool_bar['mean'] = False
            bool_bar['euclidean'] = True
            bool_bar['dba_vec'] = True
            bool_bar['dba_sub'] = True
            bool_bar['soft_dtw'] = True
        if plt_barycenters:
            bool_plots['plt_all_barycenters'], bool_plots['summary_barycenters'], = True, True
        if plt_clu:
            bool_plots['view_clusters'], bool_plots['pct_clusters'] = True, True
        train_model = self.train_clu_model(n_clusters=n_clusters)
        if self.reduced is False and self.clu_lbl is None:
            return self._dict_barycenter(
                train_model, bool_bar, bool_plots, n_clusters, max_iter, tol, verbose, gamma
            )
        if self.reduced and self.clu_lbl is None:
            return self._dict_barycenter(
                train_model, bool_bar, bool_plots, n_clusters, max_iter, tol, verbose, gamma
            )
        if self.reduced and self.clu_lbl is not None:
            print('here')
            dict_bar_labels = {}
            for i in self.clu_lbl.unique():
                dict_bar_labels[i] = self._dict_barycenter(
                    train_model[i], bool_bar, bool_plots, n_clusters, max_iter, tol, verbose, gamma
                )
            return dict_bar_labels
        if self.reduced is False and self.clu_lbl is not None:
            dict_bar_labels = {}
            for i in self.clu_lbl.unique():
                dict_bar_labels[i] = self._dict_barycenter(
                    train_model[i], bool_bar, bool_plots, n_clusters, max_iter, tol, verbose, gamma
                )
            return dict_bar_labels

        #################################################
        if self.clu_lbl is not None:
            dict_bar_labels = {}
            for i in self.clu_lbl.unique():
                dict_bar_labels[i] = self._dict_barycenter(
                    train_model[i], bool_bar, bool_plots, n_clusters, max_iter, tol, verbose, gamma
                )
            return dict_bar_labels
        else:
            return self._dict_barycenter(
                train_model, bool_bar, bool_plots, n_clusters, max_iter, tol, verbose, gamma
            )
        return None

    def _dict_barycenter(self, train_model, bool_bar, bool_plots, n_clusters, max_iter, tol, verbose, gamma):

        dict_bar = {
            'dataset': train_model['dict_scaler']['dataset'],
            'clusters': [x.T for x in train_model['ts_clustered']]
        }
        # Find barycentre's
        bar = barycenter(train_model['ts_clustered'], max_iter, tol, verbose, gamma)
        if self.model in clu_no_centroid:
            pass
        else:
            if bool_bar['centroid']:
                dict_bar['centroid'] = bar.centroid(
                    train_model['clu_model'],
                    train_model['ts_scaler'],
                    train_model['data']
                )
        if bool_bar['mean']:
            dict_bar['mean'] = bar.mean()
        if bool_bar['euclidean']:
            dict_bar['euclidean'] = bar.euclidean()
        if bool_bar['dba_vec']:
            dict_bar['dba_vectorized'] = bar.dba_vectorized()
        if bool_bar['dba_sub']:
            dict_bar['dba_subgradient'] = bar.dba_subgradient()
        if bool_bar['soft_dtw']:
            dict_bar['soft_dtw'] = bar.soft_dtw()
        for key, values_scales in dict_bar.items():
            if key == 'dataset':
                dict_bar['dataset'] = pd.DataFrame(
                    self._inverse_transform(values_scales.values.T, train_model['dict_scaler']).T,
                    columns=train_model['dict_scaler']['dataset'].columns,
                    index=train_model['dict_scaler']['dataset'].index
                )
            else:
                for x in values_scales:
                    self._inverse_transform(x, train_model['dict_scaler'])
        if bool_plots['summary_barycenters']:
            fx_clu_plots.all_barycenter(True, False, bool_plots['plot_n_cluster'], gamma, dict_bar)
        if bool_plots['plt_all_barycenters']:
            fx_clu_plots.all_barycenter(False, True, bool_plots['plot_n_cluster'], gamma, dict_bar)
        if bool_plots['view_clusters']:
            self._view_cluster(n_clusters, dict_bar, gamma)
        if bool_plots['pct_clusters']:
            fx_clu_plots.pct_clusters(train_model['cluster_labels'])
        dict_bar['dataset']['cluster'] = train_model['cluster_labels']
        dict_bar['dataset']['user'] = train_model['dict_scaler']['user']
        fx_clu_plots._clu_barplot(dict_bar)
        return dict_bar

    def _get_hyperparameters(
            self,
            data: np.array = None
    ):
        if data is None:
            dict_scaler = self.transformer_dataset()
            data = dict_scaler['dataset'].values
        if self.type_dr is not None:
            data = self.dimensionality_reduction().fit_transform(data)
        model = NearestNeighbors(n_neighbors=2)  # creating an object of the NearestNeighbors class
        train_model = model.fit(data)  # fitting the data to the object
        distances, indices = train_model.kneighbors(data)  # finding the nearest neighbours
        # Sort and plot the distances results
        distances = np.sort(distances, axis=0)  # sorting the distances
        distances = distances[:, 1]  # taking the second column of the sorted distances
        plt.rcParams['figure.figsize'] = (5, 3)  # setting the figure size
        plt.plot(distances)  # plotting the distances
        plt.show()  # showing the plot

    def reduce_data(self):
        df_reduce = pd.DataFrame()

    def optimal_number_of_clusters(
            self,
            data: np.array = None,
            max_clusters: int = 10,
            max_quantile: float = 1.0,
            step: float = 0.05,
            plot: bool = True,
            n_clusters: int = 3,
            n_components: int = 3,
            quantile: float = 0.13,
            n_samples: int = None,
            eps: float = 1.1,
            min_samples=None,
            threshold: float = 1.5,
            branching_factor: int = 50,
            damping: float = 0.90,
            sigma: float = 0.5,
            lr: float = 0.5,
            num_iteration: int = 500,
            points_2d: bool = False,
            view_clu: bool = False,

    ):
        """
        Runs KMeans n times (according to max_cluster range)

        datasets: pd.DataFrame or np.array
            Time Series Data
        max_clusters: int
            Number of different clusters for KMeans algorithm
        metric: str
            Distance metric between the observations
        Returns:
        -------
        None
        """

        if self.reduced is False and self.clu_lbl is None:
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_scaler = self._scaler_data(dict_sce)
            ts_scaler = dict_scaler['dataset'].values
            # Dimensional reduction
            if self.type_dr is not None:
                data = self.dimensionality_reduction().fit_transform(ts_scaler)
            else:
                data = ts_scaler
        if self.reduced and self.clu_lbl is None:
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_scaler = self._scaler_data(dict_sce)
            ts_scaler = dict_scaler['dataset'].values
            # Dimensional reduction
            if self.type_dr is not None:
                data = self.dimensionality_reduction().fit_transform(ts_scaler)
            else:
                data = ts_scaler
        if self.reduced and self.clu_lbl is not None:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
        if self.reduced is False and self.clu_lbl is not None:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
        if self.clu_lbl is not None and self.reduced:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
        if self.clu_lbl is not None:
            models_with_labels = {}
            dict_sce = self.reshape_to_clu_atypical_mean()
            for i in self.clu_lbl.unique():
                dict_scaler = self._scaler_data(dict_sce, i)
                ts_scaler = dict_scaler['dataset'].values
                if self.type_dr is not None:
                    data = self.dimensionality_reduction().fit_transform(ts_scaler)
                else:
                    data = ts_scaler
        else:
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_scaler = self._scaler_data(dict_sce)
            ts_scaler = dict_scaler['dataset'].values
            # Dimensional reduction
            if self.type_dr is not None:
                data = self.dimensionality_reduction().fit_transform(ts_scaler)
            else:
                data = ts_scaler

        '''
        if data is None:
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_scaler = self._scaler_data(dict_sce)
            data = dict_scaler['dataset'].values
        if self.type_dr is not None:
            data = self.dimensionality_reduction().fit_transform(data)
        '''
        parameters = {
            'type_dr': self.type_dr,
            'model': self.model,
            'metric': self.metric,
            'n_jobs': self.n_jobs,
            'max_iter': self.max_iter,
            'seed': self.seed,
            'n_clusters': n_clusters,
            'n_components': n_components,
            'quantile': quantile,
            'n_samples': n_samples,
            'eps': eps,
            'min_samples': min_samples,
            'threshold': threshold,
            'branching_factor': branching_factor,
            'damping': damping,
            'sigma': sigma,
            'lr': lr,
            'num_iteration': num_iteration,
            'points_2d': points_2d,
            'view_clu': view_clu
        }

        # list empty for metrics
        distortions = []
        silhouette = []
        davies_bouldin = []
        calinski_harabasz = []
        range_aux = []
        # Number the test
        if self.model == 'MeanShift':
            clusters_range = np.arange(0.1, max_quantile, step)
            for k in tqdm(clusters_range):
                parameters['quantile'] = k
                clu = _choice_clu_model(parameters)
                clu_model = clu.check_model(data, ts_scaler, dict_scaler)
                clu_labels = clu_model.fit_predict(data)
                # distortions.append(clu_model.inertia_)
                if k < 1:
                    if len(np.unique(clu_labels)) > 1:
                        silhouette.append(silhouette_score(data, clu_labels))
                        davies_bouldin.append(davies_bouldin_score(data, clu_labels))
                        calinski_harabasz.append(calinski_harabasz_score(data, clu_labels))
                        range_aux.append(k)
            clusters_range = range_aux
        elif self.model in model_with_eps:
            self._get_hyperparameters()
            return None
        else:
            clusters_range = range(1, max_clusters + 1)
            for k in tqdm(clusters_range):
                parameters['n_clusters'] = k
                clu = _choice_clu_model(parameters)
                clu_model = clu.check_model(data, ts_scaler, dict_scaler)
                clu_labels = clu_model['clu_model'].fit_predict(data)
                if k == 1:
                    distortions.append(clu_model['clu_model'].inertia_)
                    silhouette.append(np.NaN)
                    davies_bouldin.append(np.NaN)
                    calinski_harabasz.append(np.NaN)
                    range_aux.append(k)
                if k > 1:
                    distortions.append(clu_model['clu_model'].inertia_)
                    silhouette.append(silhouette_score(data, clu_labels))
                    davies_bouldin.append(davies_bouldin_score(data, clu_labels))
                    calinski_harabasz.append(calinski_harabasz_score(data, clu_labels))
                    range_aux.append(k)
            clusters_range = range_aux

        metrics = {
            'distortions': distortions,
            'clusters_range': clusters_range,
            'silhouette': silhouette,
            'davies_bouldin': davies_bouldin,
            'calinski_harabasz': calinski_harabasz
        }

        if plot:
            self.plots.optimal_numbers_clustering(metrics, self.model)
        return metrics

        # Visualization
        fig, ax = plt.subplots()
        fig.subplots_adjust(right=0.75)
        twin1 = ax.twinx()
        twin2 = ax.twinx()
        twin3 = ax.twinx()

        # Offset the right spine of twin2.  The ticks and label have already been
        # placed on the right by twinx above.
        twin2.spines.right.set_position(("axes", 1.1))
        twin3.spines.right.set_position(("axes", 1.2))

        p1, = ax.plot(clusters_range, distortions, "bx-", label="Intra-cluster")
        p2, = twin1.plot(clusters_range[1:], silhouette, "rx-", label="Silhouette")
        p3, = twin2.plot(clusters_range[1:], davies_bouldin, "gx-", label="Davies Bouldin")
        p4, = twin3.plot(clusters_range[1:], calinski_harabasz, "yx-", label="Calinski Harabasz")

        ax.set_xlabel("Number of clusters (k)")
        ax.set_ylabel("Inertia")
        twin1.set_ylabel("Silhouette score")
        twin2.set_ylabel("Davies Bouldin score")
        twin3.set_ylabel("Alinski_Harabasz score")

        ax.yaxis.label.set_color(p1.get_color())
        twin1.yaxis.label.set_color(p2.get_color())
        twin2.yaxis.label.set_color(p3.get_color())
        twin3.yaxis.label.set_color(p4.get_color())

        tkw = dict(size=4, width=1.5)
        ax.tick_params(axis='y', colors=p1.get_color(), **tkw)
        twin1.tick_params(axis='y', colors=p2.get_color(), **tkw)
        twin2.tick_params(axis='y', colors=p3.get_color(), **tkw)
        twin3.tick_params(axis='y', colors=p4.get_color(), **tkw)
        ax.tick_params(axis='x', **tkw)

        ax.legend(handles=[p1, p2, p3, p4])
        plt.show()

    def performance_to_clu(
            self,
            dict_results,
            metrics: bool = True,
            pairplot_global: bool = False,
            violinplot_global: bool = False,
            boxplot_global: bool = False,
            boxplot_nMEAS: bool = False,
            PDC_AMI: bool = False,
            n_plots: int = 3,
            all_metrics: bool = True,
            mse: bool = None,
            mae: bool = None,
            mape: bool = None,
            m_global: bool = True,
            m_partial: bool = False
    ):
        preview = {
            'PDC_AMI': PDC_AMI,
            'n_plots': n_plots
        }
        dict_metrics = {
            'mse': mse,
            'mae': mae,
            'mape': mape,
            'case': {
                'm_global': m_global,
                'm_partial': m_partial
            },
            'plots': {
                'pairplot_global': pairplot_global,
                'boxplot_global': boxplot_global,
                'violinplot_global': violinplot_global,
                'boxplot_nMEAS': boxplot_nMEAS
            }
        }
        if all_metrics:
            dict_metrics['mse'], dict_metrics['mae'], dict_metrics['mape'] = False, False, True

        if self.reduced is False and self.clu_lbl is None:
            aux = 2
            dict_dates = self.get_summation_MEAS_Barycenter(dict_results, aux)
        if self.reduced and self.clu_lbl is None:
            aux = 2
            dataset_aux = pd.DataFrame()
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_sce.pop('means')
            df_results = dict_results['dataset'][['cluster', 'user']].reset_index(drop=True)
            df_aux = pd.DataFrame()
            for key, value in dict_sce.items():
                clu = df_results[df_results.user == key].values[0, 0]
                value.loc[:, 'cluster'] = clu
                value.loc[:, 'user'] = key
                dataset_aux = pd.concat([dataset_aux, value])
            dict_results['dataset'] = dataset_aux
            del dataset_aux
            del dict_sce
            dict_dates = self.get_summation_MEAS_Barycenter(dict_results, aux)

        if self.reduced and self.clu_lbl is not None:
            aux = 3
            dataset_aux = pd.DataFrame()
            dict_sce = self.reshape_to_clu_atypical_mean()
            dict_sce.pop('means')
            for x in self.clu_lbl.unique():
                df_results = dict_results[x]['dataset'][['cluster', 'user']].reset_index(drop=True)
                for key, value in dict_sce.items():
                    clu = df_results[df_results.user == key].values[0, 0]
                    df_aux = value[value.clu_1 == x]
                    df_aux.loc[:, 'cluster'] = clu
                    df_aux.loc[:, 'user'] = key
                    dataset_aux = pd.concat([dataset_aux, df_aux])
            dataset_aux.sort_index(inplace=True)
            dataset_aux.rename(columns={'clu_1': 'label'}, inplace=True)
            dict_results['dataset'] = dataset_aux
            del dataset_aux
            del dict_sce
            dict_dates = self.get_summation_MEAS_Barycenter(dict_results, aux)
            if preview['PDC_AMI']:
                self.plots.summation_MEAS_Barycenters(dict_dates, preview['n_plots'], self.seed)
            if metrics:
                dict_results_metrics = self.calculate_metrics(dict_dates, dict_metrics)
            return dict_results_metrics

        if self.reduced is False and self.clu_lbl is not None:
            aux = 3
            for i in self.clu_lbl.unique():
                dict_results[i]['dataset'].loc[:, 'label'] = i
            df_dataset = pd.DataFrame()
            for key, value in dict_results.items():
                df_dataset = pd.concat([df_dataset, value['dataset']])
                del value['dataset']
            df_dataset.sort_index(inplace=True)
            dict_results['dataset'] = df_dataset
            dict_dates = self.get_summation_MEAS_Barycenter(dict_results, aux)
        if preview['PDC_AMI']:
            self.plots.summation_MEAS_Barycenters(dict_dates, preview['n_plots'], self.seed)
        if metrics:
            dict_results_metrics = self.calculate_metrics(dict_dates, dict_metrics)
        return dict_results_metrics

    def get_summation_MEAS_Barycenter(self, dict_results, aux):
        dates = dict_results['dataset'].index.unique()
        dict_dates = dict()
        for d in dates:
            dict_data = dict()
            df_filter = dict_results['dataset'][dict_results['dataset'].index == d]
            meas_PDC = df_filter.iloc[:, :-aux].T.sum(axis=1).values
            dict_data['PDC'] = meas_PDC

            if self.reduced is False and self.clu_lbl is None:
                for i in dict_results.keys():
                    if i == 'dataset' or i == 'clusters':
                        pass
                    else:
                        np_aux = np.zeros((df_filter.shape[1] - aux, 1))
                        for j in df_filter.cluster:
                            np_aux += dict_results[i][j]
                        dict_data[i] = np_aux.reshape(df_filter.shape[1] - aux, )
                dict_dates[d] = dict_data
            if self.reduced and self.clu_lbl is None:
                for i in dict_results.keys():
                    if i == 'dataset' or i == 'clusters':
                        pass
                    else:
                        np_aux = np.zeros((df_filter.shape[1] - aux, 1))
                        for j in df_filter['cluster']:
                            np_aux += dict_results[i][j]
                        dict_data[i] = np_aux.reshape(df_filter.shape[1] - aux, )
                dict_dates[d] = dict_data
            if self.reduced and self.clu_lbl is not None:
                k = df_filter.label.unique()[0]
                if self.clu_lbl is not None:
                    for i in dict_results[k].keys():
                        if i == 'dataset' or i == 'clusters':
                            pass
                        else:
                            np_aux = np.zeros((df_filter.shape[1] - aux, 1))
                            for j in df_filter.cluster:
                                np_aux += dict_results[k][i][j]
                            dict_data[i] = np_aux.reshape(df_filter.shape[1] - aux, )
                    dict_dates[d] = dict_data

            if self.reduced is False and self.clu_lbl is not None:
                k = df_filter.label.unique()[0]
                if self.clu_lbl is not None:
                    for i in dict_results[k].keys():
                        if i == 'dataset' or i == 'clusters':
                            pass
                        else:
                            np_aux = np.zeros((df_filter.shape[1] - aux, 1))
                            for j in df_filter.cluster:
                                np_aux += dict_results[k][i][j]
                            dict_data[i] = np_aux.reshape(df_filter.shape[1] - aux, )
                    dict_dates[d] = dict_data
        return dict_dates

    def calculate_metrics(self, dict_dates, dict_metrics):
        df_global, df_partial = None, None
        if dict_metrics['case']['m_global']:
            df_global = pd.DataFrame()
        if dict_metrics['case']['m_partial']:
            df_partial = pd.DataFrame()
        for date in dict_dates.keys():
            for key, values in dict_dates[date].items():
                true = dict_dates[date]['PDC']
                if key == 'PDC':
                    pass
                else:
                    predict = values
                    if dict_metrics['case']['m_global']:
                        cols = ['model', 'date']
                        list_global = [key, date]
                        if dict_metrics['mse']:
                            list_global.append(mean_squared_error(true, predict.reshape(-1), squared=False))
                            cols.append('MSE')
                        if dict_metrics['mae']:
                            list_global.append(mean_absolute_error(true, predict.reshape(-1)))
                            cols.append('MAE')
                        if dict_metrics['mape']:
                            list_global.append(mean_absolute_percentage_error(true, predict.reshape(-1)))
                            cols.append('MAPE')
                        df_aux = pd.DataFrame(np.array(list_global).reshape(1, len(cols)), columns=cols)
                        df_global = pd.concat([df_global, df_aux], axis=0)
                    if dict_metrics['case']['m_partial']:
                        aux = np.concatenate((true.reshape(-1, 1), predict.reshape(-1).reshape(-1, 1)), axis=1)
                        for j in range(0, aux.shape[0]):
                            cols = ['Nro', 'model', 'date']
                            m_ind = [j, key, date]
                            if dict_metrics['mse']:
                                m_ind.append(
                                    mean_squared_error(
                                        aux[j][1].reshape(1, -1), squared=False
                                    )
                                )
                                cols.append('MSE')
                            if dict_metrics['mae']:
                                m_ind.append(
                                    mean_absolute_error(
                                        aux[j][0].reshape(1, -1),
                                        aux[j][1].reshape(1, -1)
                                    )
                                )
                                cols.append('MAE')
                            if dict_metrics['mape']:
                                m_ind.append(
                                    mean_absolute_percentage_error(
                                        aux[j][0].reshape(1, -1),
                                        aux[j][1].reshape(1, -1)
                                    )
                                )
                                cols.append('MAPE')
                            df_aux2 = pd.DataFrame(np.array(m_ind).reshape(1, len(cols)), columns=cols)
                            df_partial = pd.concat([df_partial, df_aux2], axis=0)
                            m_ind.clear(), cols.clear()
        metrics_aux = {}
        if dict_metrics['case']['m_global']:
            df_global.reset_index(inplace=True, drop=True)
            df_global.set_index('date', inplace=True, drop=True)
            df_global.index = pd.to_datetime(df_global.index)
            for i in df_global.columns[1:]:
                df_global[i] = df_global[i].astype(float)
            metrics_aux['global'] = df_global
        if dict_metrics['case']['m_partial']:
            df_partial.reset_index(inplace=True, drop=True)
            for i in df_partial.columns[3:]:
                df_partial[i] = df_partial[i].astype(float)
            metrics_aux['partial'] = df_partial

        if dict_metrics['plots']['pairplot_global']:
            sns.pairplot(df_global, hue='model')
            # sns.scatterplot(data=df_metric, x=df_metric.index, y='MAE', hue='model')
            plt.legend()
            plt.show()
        if dict_metrics['plots']['boxplot_global']:
            self.plots.plt_global(metrics_aux, True, False)
        if dict_metrics['plots']['violinplot_global']:
            self.plots.plt_global(metrics_aux, False, True)
        if dict_metrics['plots']['boxplot_nMEAS']:
            self.plots.plt_partial(metrics_aux, True, False)
        return metrics_aux

    def dimensionality_reduction(self):
        if self.type_dr == 'tsne':
            return TSNE(
                n_components=self.n_components,
                init='pca',
                random_state=self.seed
            )
        if self.type_dr == 'mds':
            return MDS(
                n_components=self.n_components,
                n_init=3,
                max_iter=100,
                random_state=self.seed
            )

    # Visualization for obtained clusters

    def _view_cluster(
            self,
            n_clusters: int = None,
            dict_bar: dict = None,
            gamma: float = None
    ):
        fig_3, ax = plt.subplots(nrows=n_clusters, ncols=len(dict_bar.keys()) - 2, sharex=True)
        n = 0
        for key, value in dict_bar.items():
            if key == 'dataset' or key == 'clusters':
                pass
            else:
                for index, series in enumerate(dict_bar[key]):
                    ax[index, n].plot(dict_bar['clusters'][index], "k-", alpha=0.15)
                    ax[index, n].plot(series, c='r')
                    if n == 0:
                        ax[index, n].set_ylabel(f'Cluster {index}')
                    if index == n_clusters - 1:
                        ax[index, n].set_xlabel('Number of measurements')
                    if index == 0:
                        myTitle = f"Cluster with {set_tittle_cluster(key, gamma)}"
                        ax[index, n].set_title(myTitle, loc='center', wrap=True)
                n += 1
        plt.legend()
        fig_3.suptitle(f"Nro. {n_clusters} Clusters\nwith {self.type_dr} dimentional reduction ")
        plt.show()

    def clu_model_2d(self, model):
        # Visualization for obtained clusters
        u_labels = np.unique(model['cluster_labels'])
        # Centroids Visualization
        plt.figure(figsize=(16, 10))
        if self.model in clu_no_centroid:
            pass
        else:
            centroids = model['clu_model'].cluster_centers_
            plt.scatter(centroids[:, 0], centroids[:, 1], s=150, color='r', marker="x")
        # Downsize the datasets into 2D
        if model['data'].shape[1] > 2:
            self.type_dr = 'tsne'
            data = self.dimensionality_reduction().fit_transform(model['data'])
            for u_label in u_labels:
                cluster_points = data[(model['cluster_labels'] == u_label)]
                plt.scatter(cluster_points[:, 0], cluster_points[:, 1], label=u_label)
        else:
            for u_label in u_labels:
                cluster_points = model['data'][(model['cluster_labels'] == u_label)]
                plt.scatter(cluster_points[:, 0], cluster_points[:, 1], label=u_label)
        plt.title('Clustered Data')
        plt.xlabel("Feature space for the 1st feature")
        plt.ylabel("Feature space for the 2nd feature")
        plt.grid(True)
        plt.legend(title='Cluster Labels')
        plt.show()

    def plot_2D_clusters(
            self,
            cluster_model,
            data=None
    ):
        """
        Plots clusters obtained by clustering model

        datasets: pd.DataFrame or np.array
            Time Series Data
        cluster_model: Class
            Clustering algorithm
        dim_red_algo: Class
            Dimensionality reduction algorithm (e.g. TSNE/PCA/MDS...)
        Returns:
        -------
        None
        """
        if data is None:
            dict_scaler = self.transformer_dataset()
            data = dict_scaler['dataset'].values
        if self.type_dr is not None:
            dim_red_algo = self.dimensionality_reduction()
            data = dim_red_algo.fit_transform(data)

        cluster_labels = cluster_model.fit_predict(data)
        centroids = cluster_model.cluster_centers_
        u_labels = np.unique(cluster_labels)

        # Centroids Visualization
        plt.figure(figsize=(16, 10))
        plt.scatter(centroids[:, 0], centroids[:, 1], s=150, color='r', marker="x")

        # Downsize the datasets into 2D
        if data.shape[1] > 2:
            self.type_dr = 'tsne'
            data = self.dimensionality_reduction().fit_transform(data)
            for u_label in u_labels:
                cluster_points = data[(cluster_labels == u_label)]
                plt.scatter(cluster_points[:, 0], cluster_points[:, 1], label=u_label)
        else:
            for u_label in u_labels:
                cluster_points = data[(cluster_labels == u_label)]
                plt.scatter(cluster_points[:, 0], cluster_points[:, 1], label=u_label)

        plt.title('Clustered Data')
        plt.xlabel("Feature space for the 1st feature")
        plt.ylabel("Feature space for the 2nd feature")
        plt.grid(True)
        plt.legend(title='Cluster Labels')
        plt.show()


class fx_clu_plots:

    def __init__(self):
        return

    @staticmethod
    def _clu_barplot(dict_bar):
        df_count_clu = dict_bar['dataset'].groupby(['cluster', 'user', ]).size().unstack()
        df_count_clu = df_count_clu / df_count_clu.sum(axis=0)
        # plot df_count_clu T in barprlot stacked = True
        # df_count_clu.T.plot.bar(stacked=True, figsize=(10, 5))
        ax = df_count_clu.T.plot.bar(stacked=True, figsize=(10, 5))
        # ax.set_title("Title here!",color='black')
        ax.legend(bbox_to_anchor=(1.0, 1.0))
        ax.plot()
        plt.show()

    @staticmethod
    def pct_clusters(
            labels_clu
    ):
        labels = np.unique(labels_clu, return_counts=True)
        fig, ax = plt.subplots()
        ax.pie(labels[1], labels=labels[0], autopct='%1.1f%%', shadow=True, startangle=90)
        plt.show()

    @staticmethod
    def optimal_numbers_clustering(
            metrics: dict = None,
            model: str = None
    ):
        if metrics is None:
            print('run the function: optimal_number_of_clusters_with_kmeans')
        else:
            # Visualization
            fig, ax = plt.subplots()
            fig.subplots_adjust(right=0.75)
            twin1 = ax.twinx()
            twin2 = ax.twinx()
            twin3 = ax.twinx()

            # Offset the right spine of twin2.  The ticks and label have already been
            # placed on the right by twinx above.
            twin2.spines.right.set_position(("axes", 1.1))
            twin3.spines.right.set_position(("axes", 1.2))

            p1, = ax.plot(metrics['clusters_range'], metrics['silhouette'], "rx-", label="Silhouette")
            p2, = twin1.plot(metrics['clusters_range'], metrics['davies_bouldin'], "gx-", label="Davies Bouldin")
            p3, = twin2.plot(metrics['clusters_range'], metrics['calinski_harabasz'], "yx-", label="Calinski Harabasz")
            if len(metrics['distortions']) > 0:
                p4, = twin3.plot(metrics['clusters_range'], metrics['distortions'], "bx-", label="Intra-cluster")

            if model == 'MeanShift':
                ax.set_xlabel("Quantile (k)")
            else:
                ax.set_xlabel("Number of clusters (k)")

            ax.set_ylabel("Silhouette score")
            twin1.set_ylabel("Davies Bouldin score")
            twin2.set_ylabel("Alinski_Harabasz score")
            twin3.set_ylabel("Inertia")

            ax.yaxis.label.set_color(p1.get_color())
            twin1.yaxis.label.set_color(p2.get_color())
            twin2.yaxis.label.set_color(p3.get_color())
            if len(metrics['distortions']) > 0:
                twin3.yaxis.label.set_color(p4.get_color())

            tkw = dict(size=4, width=1.5)

            ax.tick_params(axis='y', colors=p1.get_color(), **tkw)
            twin1.tick_params(axis='y', colors=p2.get_color(), **tkw)
            twin2.tick_params(axis='y', colors=p3.get_color(), **tkw)
            if len(metrics['distortions']) > 0:
                twin3.tick_params(axis='y', colors=p4.get_color(), **tkw)
            ax.tick_params(axis='x', **tkw)

            if len(metrics['distortions']) > 0:
                ax.legend(handles=[p1, p2, p3, p4])
            else:
                ax.legend(handles=[p1, p2, p3])
            plt.show()

    @staticmethod
    def clu_model_view(
            model
    ):
        def plot_cluster_ts(current_cluster):
            """
            Plots time series in a cluster

            current_cluster: np.array
                Cluster with time series
            Returns:
            -------
            None
            """
            fig, ax = plt.subplots(
                int(np.ceil(current_cluster.shape[0] / 5)),
                5,
                figsize=(45, 3 * int(np.ceil(current_cluster.shape[0] / 5))),
                sharex=True,
                sharey=True
            )
            fig.autofmt_xdate(rotation=45)
            ax = ax.reshape(-1)
            for index, series in enumerate(current_cluster):
                ax[index].plot(series)
                plt.xticks(rotation=45)

            plt.tight_layout()
            plt.show()

        # Objects distribution in the obtained clusters
        labels = [f'Cluster_{i}' for i in range(len(model['ts_clustered']))]
        samples_in_cluster = [val.shape[0] for val in model['ts_clustered']]
        plt.figure(figsize=(16, 5))
        plt.bar(labels, samples_in_cluster)
        # plt.pie(samples_in_cluster, labels=labels, autopct='%1.1f%%', shadow=True, startangle=90)
        for cluster in range(len(model['ts_clustered'])):
            print(f"==========Cluster number: {cluster}==========")
            plot_cluster_ts(model['ts_clustered'][cluster])
        # plt.show()

    @staticmethod
    def all_barycenter(
            summary: bool = True,
            all_bar: bool = False,
            plot_n_cluster: int = 1,
            gamma: int = 1.0,
            dict_bar: dict = None
    ):
        if all_bar:
            # Plot barycenter of the cluster input for user and selected
            fig_1, ax = plt.subplots(nrows=len(dict_bar.keys()) - 2, ncols=1, sharex=True)
            n = 0
            for key, value in dict_bar.items():
                if key == 'dataset' or key == 'clusters':
                    pass
                else:
                    ax[n].plot(dict_bar['clusters'][plot_n_cluster], "k-", alpha=0.2)
                    ax[n].plot(dict_bar[key][plot_n_cluster], c='r')
                    ax[n].set_title(set_tittle_cluster(key, gamma), loc='left', y=0.85, x=0.02, fontsize='medium')
                    n += 1
            fig_1.suptitle(f'Cluster Series - Cluster No. {plot_n_cluster}')
            plt.show()

        if summary:
            fig, ax = plt.subplots(nrows=len(dict_bar['clusters']), ncols=1, sharex=True, layout='constrained')
            for i in range(len(dict_bar['clusters'])):
                ax[i].plot(dict_bar['clusters'][i], "k-", alpha=0.2)
                for key, value in dict_bar.items():
                    if (key == 'dataset') or (key == 'clusters'):
                        pass
                    else:
                        ax[i].plot(dict_bar[key][i], label=set_tittle_cluster(key, gamma))
                ax[i].set_title(f"Summary to Cluster Nro. {i}")
            labels_handles = {
                label: handle for ax in fig.axes for handle, label in zip(*ax.get_legend_handles_labels())
            }
            lines = []
            labels = []
            for ax in fig.axes:
                Line, Label = ax.get_legend_handles_labels()
                # print(Label)
                lines.extend(Line)
                labels.extend(Label)

            lines = [i for n, i in enumerate(lines) if i not in lines[:n]]
            labels = [i for n, i in enumerate(labels) if i not in labels[:n]]

            # rotating x-axis labels of last sub-plot
            plt.xticks(rotation=45)

            fig.legend(lines, labels, loc='outside right')
            plt.show()

    @staticmethod
    def summation_MEAS_Barycenters(dict_dates, preview, seed):
        import random
        random.seed(seed)
        dates_aux = random.choices(list(dict_dates.keys()), k=preview)
        for d in dates_aux:
            plt.figure(figsize=(15, 5))
            for key, values in dict_dates[d].items():
                if key == 'PDC':
                    plt.plot(values, label=key, color='k', lw=3, marker='o')
                else:
                    plt.plot(values, label=key, ls='--')
            plt.title(d)
            plt.legend(loc="best")

            plt.show()

    @staticmethod
    def plt_global(dict_metrics, boxplot: bool = None, violinplot: bool = None):
        df_global = dict_metrics['global']
        if boxplot:
            # Boxplot
            if len(df_global.columns[1:]) == 1:
                sns.boxplot(data=df_global, x='model', y=df_global.columns[1:][0], fliersize=0)
                plt.title('Boxplot for clustering')
                plt.show()
            else:
                n = len(df_global.columns[1:])
                fig, axes = plt.subplots(n, 1, sharex=True)
                n = 1
                for k in df_global.columns[1:]:
                    sns.boxplot(ax=axes[n], data=df_global, x='model', y=k, fliersize=0)
                    n += 1
                fig.suptitle('Boxplot for clustering')
                plt.show()

        if violinplot:
            # Violinplot
            # Boxplot
            if len(df_global.columns[1:]) == 1:
                sns.violinplot(data=df_global, x='model', y=df_global.columns[1:][0], fliersize=0)
                plt.title('Violinplot for clustering')
                plt.show()
            else:
                n = len(df_global.columns[1:])
                fig, axes = plt.subplots(n, 1, sharex=True)
                n = 1
                for k in df_global.columns[1:]:
                    sns.violinplot(ax=axes[n], data=df_global, x='model', y=k, fliersize=0)
                    n += 1
                fig.suptitle('Violinplot for clustering')
                plt.show()

    @staticmethod
    def plt_partial(dict_metrics, boxplot: bool = None, violinplot: bool = None):
        df_partial = dict_metrics['partial']
        if boxplot:
            fig2, axes = plt.subplots(3, 1, figsize=(18, 10))
            fig2.suptitle('Boxplot for clustering')
            sns.boxplot(ax=axes[0], data=df_partial[df_partial.model == 'mean'], x='Nro', y='MAPE', fliersize=0)
            sns.lineplot(ax=axes[1], data=df_partial[df_partial.model == 'mean'], x='Nro', y='MAPE')
            sns.catplot(ax=axes[2], data=df_partial, x='Nro', y='MAPE', hue='model', kind="point")
            # sns.boxplot(ax=axes[2], data=df_metric_min, x='model', y='MAPE')
            sns.boxplot()


class _choice_clu_model:
    def __init__(
            self,
            parameters
    ):
        self.type_dr = parameters['type_dr']
        self.model = parameters['model']
        self.metric = parameters['metric']
        self.n_jobs = parameters['n_jobs']
        self.max_iter = parameters['max_iter']
        self.seed = parameters['seed']
        self.n_clusters = parameters['n_clusters']
        self.n_components = parameters['n_components']
        self.quantile = parameters['quantile']
        self.n_samples = parameters['n_samples']
        self.eps = parameters['eps']
        self.min_samples = parameters['min_samples']
        self.threshold = parameters['threshold']
        self.branching_factor = parameters['branching_factor']
        self.damping = parameters['damping']
        self.sigma = parameters['sigma']
        self.lr = parameters['lr']
        self.num_iteration = parameters['num_iteration']
        self.points_2d = parameters['points_2d']
        self.view_clu = parameters['view_clu']

    def check_model(self, data, ts_scaler, dict_scaler):
        # Select model of clustering
        if self.min_samples is None:
            self.min_samples = 2 * data.shape[1]
        if self.model == 'SOM':
            # Initialization and training
            som_shape = (1, self.n_clusters)
            som = MiniSom(
                som_shape[0],
                som_shape[1],
                data.shape[1],
                sigma=self.sigma,
                learning_rate=self.lr,
                neighborhood_function='gaussian',
                random_seed=10
            )
            som.train_batch(data, num_iteration=self.num_iteration, verbose=True)
            # each neuron represents a cluster
            winner_coordinates = np.array([som.winner(x) for x in data]).T
            # with np.ravel_multi_index we convert the bidimensional
            # coordinates to a monodimensional index
            cluster_labels = np.ravel_multi_index(winner_coordinates, som_shape)
            ts_clustered = [ts_scaler[(cluster_labels == label), :] for label in np.unique(cluster_labels)]
            train_model = {
                'clu_model': som,
                'ts_clustered': ts_clustered,
                'cluster_labels': cluster_labels,
                'data': data,
                'ts_scaler': ts_scaler,
                'dict_scaler': dict_scaler
            }
            train_model['dataset'] = train_model['dict_scaler']['dataset']
            #train_model['dataset']['cluster'] = train_model['cluster_labels']
            #train_model['dataset']['user'] = train_model['dict_scaler']['user']
            if self.points_2d:
                # plotting the clusters using the first 2 dimensions of the data
                for c in np.unique(cluster_labels):
                    plt.scatter(
                        data[cluster_labels == c, 0],
                        data[cluster_labels == c, 1],
                        label='cluster=' + str(c),
                        alpha=.7
                    )
                # plotting centroids
                for centroid in som.get_weights():
                    plt.scatter(
                        centroid[:, 0],
                        centroid[:, 1],
                        marker='x',
                        s=80,
                        linewidths=5,
                        color='k',
                        label='centroid'
                    )
                plt.legend()
            if self.view_clu:
                fx_clu_plots.clu_model_view(train_model)
            return train_model

        clu_model = self._select_model(data)
        cluster_labels = clu_model.fit_predict(data)
        ts_clustered = [ts_scaler[(cluster_labels == label), :] for label in np.unique(cluster_labels)]
        train_model = {
            'clu_model': clu_model,
            'ts_clustered': ts_clustered,
            'cluster_labels': cluster_labels,
            'data': data,
            'ts_scaler': ts_scaler,
            'dict_scaler': dict_scaler
        }
        train_model['dataset'] = train_model['dict_scaler']['dataset']
        #train_model['dataset']['cluster'] = train_model['cluster_labels']
        #train_model['dataset']['user'] = train_model['dict_scaler']['user']
        # train_model['clusters'] = [x.T for x in train_model['ts_clustered']]
        '''
        if self.points_2d:
            self.clu_model_2d(train_model)
        '''
        if self.view_clu:
            fx_clu_plots.clu_model_view(train_model)
        return train_model

    def _select_model(self, data):
        # Clustering model
        if self.model == 'KMeans':
            clu_model = TimeSeriesKMeans(
                n_clusters=self.n_clusters,
                metric=self.metric,
                n_jobs=self.n_jobs,
                max_iter=self.max_iter,
                random_state=self.seed
            )
            return clu_model
        elif self.model == 'MiniBatchKMeans':
            clu_model = MiniBatchKMeans(
                n_clusters=self.n_clusters,
                max_iter=self.max_iter,
                random_state=self.seed
            )
            return clu_model
        elif self.model == 'MeanShift':
            return MeanShift(
                bandwidth=estimate_bandwidth(
                    data,
                    quantile=self.quantiles,
                    n_samples=self.n_sampless
                )
            )
        elif self.model == 'DBSCAN':
            return DBSCAN(eps=self.eps, min_samples=self.min_samples)
        elif self.model == 'Birch':
            return Birch(
                n_clusters=self.n_clusters,
                threshold=self.threshold,
                branching_factor=self.branching_factor
            )
        elif self.model == 'GaussianMixture':
            return GaussianMixture(
                n_components=self.n_components
            )
        elif self.model == 'OPTICS':
            return OPTICS(
                eps=self.eps,
                min_samples=self.min_samples
            )
        elif self.model == 'AffinityPropagation':
            return AffinityPropagation(
                damping=self.damping
            )
        elif self.model == 'Agglomerative':
            return AgglomerativeClustering(
                n_clusters=self.n_clusters
            )
        elif self.model == 'Spectral':
            return SpectralClustering(
                n_clusters=self.n_clusters
            )


class barycenter:

    def __init__(self, ts_clustered, max_iter, tol, verbose, gamma):
        self.ts_clustered = ts_clustered
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.gamma = gamma

    def centroid(self, clu_model, ts_scaler, data):
        # Cluster Centroid
        closest_clusters_index = [
            np.argmin(
                [np.linalg.norm(cluster_center - point, ord=2) for point in data]
            ) for cluster_center in clu_model.cluster_centers_
        ]
        closest_ts = ts_scaler[closest_clusters_index, :]
        # Centroids
        all_centroids = [np.array(x).reshape(np.array(x).shape[0], 1) for x in closest_ts.tolist()]
        return all_centroids

    def mean(self):
        # Means to cluster's
        all_meas = [
            np.mean(np.array(x), axis=0).reshape(np.mean(np.array(x), axis=0).shape[0], 1) for x in self.ts_clustered
        ]
        return all_meas

    def euclidean(self):
        euclidean = [euclidean_barycenter(cluster_series) for cluster_series in self.ts_clustered]
        return euclidean

    def dba_vectorized(self):
        dba_ts = [
            dtw_barycenter_averaging(
                cluster_series,
                max_iter=self.max_iter,
                tol=self.tol,
                verbose=self.verbose
            ) for cluster_series in self.ts_clustered
        ]
        return dba_ts

    def dba_subgradient(self):
        dba_sub = [
            dtw_barycenter_averaging_subgradient(
                cluster_series,
                max_iter=self.max_iter,
                tol=self.tol,
                verbose=self.verbose
            ) for cluster_series in self.ts_clustered
        ]
        return dba_sub

    def soft_dtw(self):
        dba_sotf = [
            softdtw_barycenter(
                cluster_series,
                gamma=self.gamma,
                max_iter=self.max_iter,
                tol=self.tol
            ) for cluster_series in self.ts_clustered
        ]
        return dba_sotf



import tensorflow as tf

import pandas as pd
import numpy as np
from .dtw_funcs import slow_dtw_distance
from dtaidistance import dtw
from sklearn.metrics import mean_squared_error, r2_score
import matplotlib.pyplot as plt

tkm = tf.keras.metrics


def _tf_evaluate_forecast_global(y_test_inverse, yhat_inverse):
    _rmse = tkm.RootMeanSquaredError()
    _mae = tkm.MeanAbsoluteError()
    _mape = tkm.MeanAbsolutePercentageError()

    mae = _mae(y_test_inverse, yhat_inverse)
    print('MAE:', float(mae))
    rmse = _rmse(y_test_inverse, yhat_inverse)
    print('RMSE:', float(rmse))
    mape = _mape(y_test_inverse, yhat_inverse)
    print('MAPE:', float(mape))


def _metrics_3D(
        y_colname,
        n_output,
        y_test,
        yhat,
        name_model=None,
        path_save: str = None,
        scaler=None
):
    import pandas as pd
    from sklearn.metrics import mean_squared_error, mean_absolute_error

    if scaler != None:
        d1, d2, d3 = y_test.shape[0], y_test.shape[1], y_test.shape[2]

        y_test = y_test.reshape_out(d1 * d2, d3)
        yhat = yhat.reshape_out(d1 * d2, d3)
        y_test = scaler.inverse_transform(y_test)
        yhat = scaler.inverse_transform(yhat)

        y_test = y_test.reshape_out(d1, d2, d3)
        yhat = yhat.reshape_out(d1, d2, d3)

    dict_values = {}
    for a in range(len(y_colname)):
        col_true, col_pred, col_mae, col_mse = [], [], [], []
        for i in range(n_output):
            col_true.append(f'True_{i}')
            col_pred.append(f'Pred_{i}')
            col_mae.append(f'MAE_{i}')
            col_mse.append(f'MSE_{i}')
        col_value = col_true + col_pred + col_mae + col_mse
        df_value = pd.DataFrame(
            columns=col_value,
            index=range(y_test.shape[0])
        )

        for jj in range(y_test.shape[0]):
            list_true, list_pred, list_mae, list_mse = [], [], [], []
            for m in range(n_output):
                true, pred = y_test[jj][:, a: a + 1][m], yhat[jj][:, a: a + 1][m]
                list_true.append(true[0])
                list_pred.append(pred[0])
                list_mae.append(mean_absolute_error(true, pred))
                list_mse.append(mean_squared_error(true, pred))
            list_value = list_true + list_pred + list_mae + list_mse
            df_value.iloc[jj] = list_value

        if path_save != None:
            key = y_colname[a]
            df_value.to_csv(
                f'{path_save}/{name_model}_{key}.csv',
                index=False
            )
        dict_values[y_colname[a]] = df_value

    list_mae_vec = []
    col_3d = ['MAE_3D', 'MSE_3D']
    col_2d = []
    for i in range(len(y_colname)):
        col_2d.append(f'MAE_2D_V{i}')
        col_2d.append(f'MSE_2D_V{i}')

    col_3d_2d = col_3d + col_2d
    df_value_3d = pd.DataFrame(
        columns=col_3d_2d,
        index=range(y_test.shape[0])
    )
    for ii in range(y_test.shape[0]):
        list_mae_3d, list_mse_3d = [], []
        list_mae_3d.append(mean_absolute_error(y_test[ii], yhat[ii]))
        list_mse_3d.append(mean_squared_error(y_test[ii], yhat[ii]))

        list_mae_2d, list_mse_2d = [], []
        for a in range(len(y_colname)):
            true, pred = y_test[ii][:, a: a + 1], yhat[ii][:, a: a + 1]
            list_mae_2d.append(mean_absolute_error(true, pred))
            list_mse_2d.append(mean_squared_error(true, pred))
        list_value_3d_2d = list_mae_3d + list_mse_3d + list_mae_2d + list_mse_2d

        df_value_3d.iloc[ii] = list_value_3d_2d

    if path_save != None:
        df_value.to_csv(
            f'{path_save}/{name_model}_3D.csv',
            index=False
        )
    return dict_values, df_value_3d


def _plot_boxplot_n_steps_future(dict_values_2D, df_values_3D):
    import matplotlib.pyplot as plt
    import seaborn as sns
    import re
    metrics = ['MAE', 'MSE']
    for i in metrics:
        filter = [x for x in list(df_values_3D.columns) if re.match(i, x)]
        df_aux = df_values_3D[filter]
        fig, ax = plt.subplots(figsize=(10, 5))
        boxplot = sns.boxplot(data=df_aux.astype(float), showfliers=False)
        # boxplot.axes.set_title("MAE en 3D y 2D", fontsize=16)
        boxplot.set_ylabel(i, fontsize=14)
        # boxplot.set_xlabel("Columnas", fontsize=14)
        plt.show()

    for k in metrics:
        i = 0
        fig, axes = plt.subplots(1, 4, figsize=(12, 3), constrained_layout=True)
        for x in list(dict_values_2D.resolution()):
            filter = [x for x in dict_values_2D[x].columns if re.match(k, x)]
            df_aux = dict_values_2D[x][filter]
            fig = sns.boxplot(data=df_aux, ax=axes[i], showfliers=False)
            fig.axes.set_title(x)
            fig.axes.set_ylabel(k)
            i += 1
        plt.show()


def _R_squared(y_true, y_pred):
    """
    R^2 (coefficient of determination) regression score function.
    Best possible score is 1.0, lower values are worse.
    Args:
        y_true ([np.array]): test samples
        y_pred ([np.array]): predicted samples
    Returns:
        [float]: R2
    """
    SS_res = tf.reduce_sum(tf.square(y_true - y_pred), axis=-1)
    SS_tot = tf.reduce_sum(tf.square(y_true - tf.reduce_mean(y_true, axis=-1)), axis=-1)
    return 1 - SS_res / (SS_tot + tf.keras.backend.epsilon())


def _metrics_n_steps_for_future(y_test_inverse, yhat_inverse, config, view=True, filepath=None):
    import matplotlib.pyplot as plt
    _rmse = tkm.RootMeanSquaredError()
    _mape = tkm.MeanAbsolutePercentageError()
    _mae = tkm.MeanAbsoluteError()

    # yhat_inverse_time_step and y_test_inverse_time_step are both same dimension.
    time_step_list_yhat = [[] for i in range(config['n_future'])]
    time_step_list_y_test = [[] for i in range(config['n_future'])]
    for i in range(0, yhat_inverse.shape[0]):
        for j in range(0, yhat_inverse.shape[1]):
            time_step_list_yhat[j].append(list(yhat_inverse[i][j]))
            time_step_list_y_test[j].append(list(y_test_inverse[i][j]))
    yhat_time_step = np.array(time_step_list_yhat)
    yhat_time_step = yhat_time_step.reshape(yhat_time_step.shape[0], -1)
    y_test_time_step = np.array(time_step_list_y_test)
    y_test_time_step = y_test_time_step.reshape(y_test_time_step.shape[0], -1)
    # plotting
    mae_list, r2_list, rmse_list = [], [], []
    for i in range(0, config['n_future']):
        mae = _mae(y_test_time_step[i], yhat_time_step[i])
        r2 = _R_squared(y_test_time_step[i], yhat_time_step[i])
        rmse = _rmse(y_test_time_step[i], yhat_time_step[i])
        mae_list.append(mae)
        r2_list.append(r2)
        rmse_list.append(rmse)

    list_name = ['MAE', 'R2', 'RMSE']
    list_results = [mae_list, r2_list, rmse_list]

    dict_results = dict(zip(list_name, list_results))
    fig, axs = plt.subplots(3, 1, layout=None)
    aux = 0
    for key, values in dict_results.items():
        axs[aux].plot(range(0, config['n_future']), values, marker='o')
        axs[aux].set(ylabel=key, xlabel='number of steps in future')
        aux += 1
    if view:
        plt.show()
    if filepath is not None:
        fig.savefig(filepath)


class calculate_metrics:

    def __init__(self):
        return

    def fit(self, y_true, y_pred, model_name='MODEL'):
        self._y_true = y_true
        self._y_pred = y_pred
        self._model_name = model_name
        # Metrics
        self.mae = True
        self.rmse = True
        self.mse = True
        self.mape = True
        self.r2 = True
        self.dwt = False

        _rmse = tkm.RootMeanSquaredError()
        _mse = tkm.MeanAbsoluteError()
        _mae = tkm.MeanAbsoluteError()
        _mape = tkm.MeanAbsolutePercentageError()
        _accuracy = tkm.Accuracy()
        list_metrics = []
        name_metrics = []
        y_true_1d = y_true.ravel()
        y_pred_1d = y_pred.ravel()
        
        if self.mae:
            mae = _mae(y_true_1d, y_pred_1d)
            list_metrics.append(float(mae))
            name_metrics.append('MAE')
        if self.rmse:
            rmse = _rmse(y_true_1d, y_pred_1d)
            list_metrics.append(float(rmse))
            name_metrics.append('RMSE')
        if self.mse:
            mse = _mse(y_true_1d, y_pred_1d)
            list_metrics.append(float(mse))
            name_metrics.append('MSE')
        if self.mape:
            mape = _mape(y_true_1d, y_pred_1d)
            list_metrics.append(float(mape))
            name_metrics.append('MAPE')
        if self.r2:
            r2 = r2_score(y_true_1d, y_pred_1d)
            list_metrics.append(r2)
            name_metrics.append('R2')
        if self.dwt:
            dtw_distance_model = dtw.distance(y_true_1d, y_pred_1d)
            list_metrics.append(float(dtw_distance_model))
            name_metrics.append('DWT')

        df = pd.DataFrame([list_metrics], columns=name_metrics)
        list_aux = []
        for _ in range(len(list_metrics)):
            list_aux.append(0)
        df_scale = pd.DataFrame([list_aux], columns=name_metrics)
        df_all = pd.concat([df_scale, df], axis=0, ignore_index=True).reset_index(drop=True)

        self._model_df = df[name_metrics]
        self._model_df['MODEL_NAME'] = str(model_name)
        self._all_df = df_all[name_metrics]
        self._rw_df = df_scale.loc[df_scale['RMSE'] > 0][name_metrics]
        self._rw_df['MODEL_NAME'] = 'RandomWalk'
        self._all_df = df_all[name_metrics]
        self._def_viz = pd.concat([self._model_df, self._rw_df], axis=0, ignore_index=True).reset_index(drop=True)

    def rename_model(self, model_name):
        self._model_name = model_name
        self._model_df['MODEL_NAME'] = str(model_name)
        self._def_viz = pd.concat([self._model_df, self._rw_df], axis=0, ignore_index=True).reset_index(drop=True)

    def get_df_viz(self):
        return self._def_viz

    def add_rdr(
            self,
            rdr_object
    ):
        self._def_viz = pd.concat(
            [self._def_viz, rdr_object._def_viz],
            axis=0, ignore_index=True
        ).reset_index(drop=True)
        self._def_viz = self._def_viz.drop_duplicates()
        return self._def_viz

    def plot_rdr_rank(self, models=list([])):
        if len(models) == 0:
            models = self._def_viz
        models = models.append(
            pd.DataFrame([[0, 0, 1.0, 'Perfect Score']], index=[max(models.index) + 1], columns=models.columns))
        model_ranking = models.sort_values(by='RdR_SCORE', ascending=True)
        model_ranking['RdR_SCORE'] = model_ranking['RdR_SCORE'] * 100
        # model_ranking = model_ranking.loc[~model_ranking['MODEL_NAME'].isin(['PERFECT_SCORE', 'WORST_SCORE'])]
        colors = []
        for value in model_ranking.loc[:, 'RdR_SCORE']:  # keys are the names of the boys
            if value < 0:
                colors.append('r')
            elif value <= np.max(model_ranking.loc[model_ranking['MODEL_NAME'] == 'RandomWalk']['RdR_SCORE']):
                colors.append('y')
            else:
                colors.append('g')

        fig, ax = plt.subplots()
        fig.set_size_inches(8, 10, forward=True)
        ax.barh(model_ranking['MODEL_NAME'], model_ranking['RdR_SCORE'], color=colors, alpha=0.65)
        ax.tick_params(axis="y", labelsize=10)
        # ax.set_xticks(rotation = 60, fontsize = 8)
        # for i in enumerate('RMSE:' + model_ranking['LB_RMSE'].round(4).astype(str)):
        #    plt.text(i + 0, str(v), fontweight='bold')
        # find the values and append to list
        totals = []
        for i in ax.patches:
            totals.append(i.get_width())
        # set individual bar lables using above list
        # total = sum(totals)
        for i in ax.patches:
            # get_width pulls left or right; get_y pushes up or down
            ax.text(i.get_width(), i.get_y() + .38, \
                    str(round(i.get_width(), 2)) + '%', fontsize=12,
                    color='black', weight="bold"
                    )
        plt.title(
            'Model Ranking based on RdR score' + '\n'
            + 'Best possible model = 100%' + '\n'
            + '0% = Naïve RandomWalk' + '\n'
            + '< 0% = Worst than Naïve RandomWalk (Red)' + '\n'
            + '> 0% = Better than Naïve RandomWalk (Green)'
        )
        plt.tight_layout()

    def plot_rdr(
            self,
            models=list([]),
            scatter_size=80,
            scatter_label=True,
            scatter_label_size=6,
            scatter_alpha=0.65,
            figsize=(10, 10)
    ):
        if len(models) == 0:
            models = self._def_viz
        #############################################################################################
        ################### PLOT MODEL VALIDATION GRAPH GRID ########################################
        #############################################################################################
        models = models.append(
            pd.DataFrame([[0, 0, 1.0, 'Perfect Score']], index=[max(models.index) + 1], columns=models.columns))

        import matplotlib.patches as mpatches
        # rectangle = [(0,0),(0,1),(1,1),(1,0)]
        fig1 = plt.figure(figsize=figsize)
        ax1 = fig1.add_subplot(111, aspect='equal')
        ax1.add_patch(
            mpatches.Rectangle(
                (0, 0),
                self._rw_df['DTW'].value[0],
                self._rw_df['RMSE'].value[0],
                alpha=0.1,
                color='green',
                linestyle='--',
                linewidth=1.5,
                edgecolor='grey'
            )
        )
        ax2 = fig1.add_subplot(111, aspect='equal')
        ax2.add_patch(
            mpatches.Rectangle(
                (0, self._rw_df['RMSE'].value[0]),
                np.max(self._def_viz['DTW']) * 1.15,
                np.max(self._def_viz['RMSE']) * 1.15,
                alpha=0.07, color='red',
                edgecolor=None, linewidth=0
            )
        )
        ax3 = fig1.add_subplot(111, aspect='equal')
        ax3.add_patch(
            mpatches.Rectangle(
                (self._rw_df['DTW'].value[0], 0),
                np.max(self._def_viz['DTW']) * 1.15,
                self._rw_df['RMSE'].value[0],
                alpha=0.07,
                color='red',
                edgecolor=None,
                linewidth=0
            )
        )
        plt.scatter(
            self._rw_df['DTW'],
            self._rw_df['RMSE'],
            s=scatter_size,
            label='Naïve Random Walk Score',
            color='red'
        )
        for i in range(0, len(models)):
            model = pd.DataFrame(models.iloc[i:i + 1, :])
            # print(model)
            # print(model.columns.tolist())
            if (
                    (
                            model['DTW'].value[0] >= self._rw_df['DTW'].value[0]
                    ) or (
                    model['RMSE'].value[0] >= self._rw_df['RMSE'].value[0]
            )
            ):
                plt.scatter(
                    model['DTW'],
                    model['RMSE'],
                    color='red',
                    s=scatter_size,
                    label=model['MODEL_NAME'],
                    alpha=scatter_alpha
                )
            else:
                plt.scatter(
                    model['DTW'],
                    model['RMSE'],
                    color='green',
                    s=scatter_size,
                    label=model['MODEL_NAME'],
                    alpha=scatter_alpha
                )
            if scatter_label:
                if len(model['MODEL_NAME'][i]) > 15:
                    model_name = model['MODEL_NAME'][i][:15] + str('[...]')
                else:
                    model_name = model['MODEL_NAME'][i]
                plt.annotate(model_name, (model['DTW'][i], model['RMSE'][i]), fontsize=scatter_label_size)

        plt.ylim(0, np.max(self._def_viz['RMSE']) * 1.15)
        plt.xlim(0, np.max(self._def_viz['DTW']) * 1.15)
        plt.xlabel(
            'Dynamic Time Warping Distance' + '\n' +
            '(Metric for time series shape similarity, prediction vs test set)',
            fontsize=15
        )
        plt.ylabel(
            'RMSE score' + '\n' +
            '(Metric for penalized prediction errors on test set)',
            fontsize=10
        )
        plt.title(
            'MODELS A PERFORMANCE COMPARISON for multistep forecast in the future' +
            '\n' +
            'TIME SERIE' +
            '\n' +
            'RED ZONES: Not better than Naïve model' +
            '\n' +
            'GREEN ZONE: Better than Naïve model'
        )
        # plt.legend(loc='upper left', fontsize = 15)
        plt.show()
    # print('TEST')
    # plt.savefig('all_best_preds_test.png', dpi = 800)
    # plt.tight_layout()



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, value_miss


class lstm2_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(lstm2_dense, self).__init__()
        if config is None:
            config = {
                'lstm': {
                    'units': 256,
                    'activation': 'relu'
                },
                'dropout': 0.2,
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.lstm1 = tkl.LSTM(units=config['lstm']['units'], return_sequences=True)
        self.drop1 = tkl.Dropout(config['dropout'])
        self.lstm2 = tkl.LSTM(units=config['lstm']['units'], activation=config['lstm']['activation'])
        self.drop2 = tkl.Dropout(config['dropout'])
        self.dense1 = tkl.Dense(n_future * n_out_ft, activation=config['dense']['activation'])
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.lstm1(x)
        x = self.drop1(x)
        x = self.lstm2(x)
        x = self.drop2(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x


def lstm2_dense_model(
        n_past=None,
        n_inp_ft=None,
        n_future=None,
        n_out_ft=None):
    input_layer = tkl.Input(shape=(n_past, n_inp_ft))
    mask = tkl.Masking(mask_value=-1)(input_layer)
    lstm1 = tkl.LSTM(units=256, return_sequences=True)(mask)
    drop1 = tkl.Dropout(0.3)(lstm1)
    lstm2 = tkl.LSTM(units=256, activation='relu')(drop1)
    drop2 = tkl.Dropout(0.3)(lstm2)
    dense = tkl.Dense(units=(n_future * n_out_ft), activation='linear')(drop2)
    out_layer = tkl.Reshape((n_future, n_out_ft))(dense)
    return tkm.Model(inputs=input_layer, outputs=out_layer)



# -*- coding: utf-8 -*-
# @Time    : 14/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, tkr
from openpy_fxts.architectures.fxts.arch_dicts import _get_dict_fxts
from openpy_fxts.architectures.layers.layers_class import dense_multi_out, dense_drop


class mlp_dense_class(tkm.Model):

    def __init__(
            self,
            n_past=None,
            n_inp_ft=None,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(mlp_dense_class, self).__init__()
        if config is None:
            config = _get_dict_fxts('MLP_Dense').config_mlp()
        self.n_past = n_past
        self.n_inp_ft = n_inp_ft
        self.n_future = n_future
        self.n_out_ft = n_out_ft
        self.config = config
        self.hidden_input = []
        for _ in range(n_inp_ft):
            self.hidden_input.append(
                tkl.Dense(
                    units=config['input_layer']['units'],
                    activation=config['input_layer']['activation'],
                    kernel_regularizer=tkr.L1L2(
                        l1=config['input_layer']['l1'],
                        l2=config['input_layer']['l2']
                    )
                )
            )
        self.hidden_dense = dense_drop(
            n_units=config['dense_layers']['units'],
            list_act=config['dense_layers']['activations'],
            l1=config['dense_layers']['l1'],
            l2=config['dense_layers']['l2'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate']
        )
        self.dense_out = dense_multi_out(n_future, n_out_ft, config['output_layer']['activation'])

    def call(self, inputs, training=True, **kwargs):
        x_input = []
        for i, (layer_input) in enumerate(self.hidden_input):
            x_input.append(layer_input(inputs[i]))
        x = tkl.concatenate(x_input)
        x = self.hidden_dense(x)
        x = self.dense_out(x)
        return x


# -*- coding: utf-8 -*-
# @Time    : 14/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkl, tkm
from openpy_fxts.architectures.layers.layers_class import type_rnn, rnn_drop, dense_multi_out, multi_rnn_drop
from openpy_fxts.architectures.mdits.arch_dicts import _get_dict_mdits


class birnn_dense_class(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(birnn_dense_class, self).__init__()
        if config is None:
            config = _get_dict_mdits(name_mdl='BiRNN_Dense').config_birnn()
        self.n_future = n_future
        self.n_out_ft = n_out_ft
        self.config = config
        self.hidden_rnn = multi_rnn_drop(
            mdl=type_rnn(config['rnn_layers']['type']),
            units=config['rnn_layers']['units'],
            activations=config['rnn_layers']['activations'],
            sequences=config['rnn_layers']['sequences'],
            bidirectional= config['rnn_layers']['bidirectional'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
            is_mask=True
        )
        if config['dense_layer']['activate']:
            self.hidden_dense = tkl.Dense(
                config['dense_layer']['units'],
                activation=config['dense_layer']['activation']
            )
        self.layer_output = dense_multi_out(n_future, n_out_ft, config['output_layer']['activation'])

    def call(self, inputs, training=True, **kwargs):
        x = self.hidden_rnn(inputs)
        if self.config['dense_layer']['activate']:
            x = self.hidden_dense(x)
        x = self.layer_output(x)
        return x


class multi_birnn_dense_class(tkm.Model):
    def __init__(
            self,
            n_past=None,
            n_inp_ft=None,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(multi_birnn_dense_class, self).__init__()
        if config is None:
            config = _get_dict_mdits('Multi_BiRNN_Dense').config_birnn()
        self.n_past = n_past
        self.n_inp_ft = n_inp_ft
        self.n_future = n_future
        self.n_out_ft = n_out_ft
        self.config = config
        mdl = type_rnn(config['multi_rnn_layers']['type'])
        self.hidden_inp = []
        dimensions = config['multi_rnn_layers']['units']
        activations = config['multi_rnn_layers']['activations']
        sequences = config['multi_rnn_layers']['sequences']
        self.hidden_rnn_dict = dict()
        for i in range(n_inp_ft):
            dict_rnn = {}
            for j, (unit, act, seq) in enumerate(zip(dimensions, activations, sequences)):
                dict_rnn['' + str(j)] = rnn_drop(
                    mdl=mdl,
                    n_units=unit,
                    activation=act,
                    return_seq=seq,
                    bidirectional=config['multi_rnn_layers']['bidirectional']
                )
            self.hidden_rnn_dict['' + str(i)] = dict_rnn
        self.concat = tkl.Concatenate(axis=1)
        if config['dropout']['activate']:
            self.drop1 = tkl.Dropout(config['dropout']['rate'])
        self.hidden_rnn = mdl(
            units=config['rnn_layer']['units'],
            activation=config['rnn_layer']['activation'],
            return_sequences=config['rnn_layer']['sequences']
        )
        if config['dropout']['activate']:
            self.drop2 = tkl.Dropout(config['dropout']['rate'])
        if config['dense_layer']['activate']:
            self.hidden_dense = tkl.Dense(
                config['dense_layer']['units'],
                activation=config['dense_layer']['activation']
            )
        self.dense_out = dense_multi_out(n_future, n_out_ft, config['output_layer']['activation'])

    def call(self, inputs, training=True, **kwargs):
        head_list = []
        for _, dict_aux in self.hidden_rnn_dict.items():
            for i, (_, layer_rnn) in enumerate(dict_aux.items()):
                if i == 0:
                    y = layer_rnn(inputs)
                else:
                    y = layer_rnn(y)
            head_list.append(y)
        x = self.concat(head_list)
        x = tkl.Reshape((head_list[0].shape[1], self.n_inp_ft))(x)
        if self.config['dropout']['activate']:
            x = self.drop1(x)
        x = self.hidden_rnn(x)
        if self.config['dropout']['activate']:
            x = self.drop2(x)
        if self.config['dense_layer']['activate']:
            x = self.hidden_dense(x)
        x = self.dense_out(x)
        return x





# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, value_miss,  reset_seeds, seed_value

reset_seeds(seed_value)


def rnn_dense(n_past, n_inp_ft, n_future, n_out_ft, config=None, drop=False, type: str ='GRU'):
    if type == 'RNN':
        mdl = tkl.RNN
    if type == 'GRU':
        mdl = tkl.GRU
    if type == 'LSTM':
        mdl = tkl.LSTM
    if config is None:
        config = {
            'layer_config': {
                'hidden_dimensions': [256, 256],
                'hidden_activations': ['relu', 'relu'],
                'hidden_sequences': [True, False]
            },
            'dropout': 0.3,
            'layer_out': 'relu'
        }
    dimensions = config['layer_config']['hidden_dimensions']
    activations = config['layer_config']['hidden_activations']
    sequences = config['layer_config']['hidden_sequences']
    input_train = tkl.Input(shape=(n_past, n_inp_ft))
    x = tkl.Masking(mask_value=value_miss)(input_train)
    for i, (unit, act, seq) in enumerate(zip(dimensions, activations, sequences)):
        x = mdl(units=unit, activation=act, return_sequences=seq)(x)
        if drop:
            x = tkl.Dropout(config['dropout'])(x)
    x = tkl.Dense(n_future * n_out_ft, activation=config['layer_out'])(x)
    x = tkl.Reshape((n_future, n_out_ft))(x)
    model = tkm.Model(inputs=input_train, outputs=x, name='GRU_Dense')
    return model



# -*- coding: utf-8 -*-
# @Time    : 29/05/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import re


def metrics_3D(
        config,
        valid_y,
        yhat,
        name_model=None,
        path_save: str = None,
        scaler=None
):
    if scaler != None:
        d1, d2, d3 = valid_y.shape[0], valid_y.shape[1], valid_y.shape[2]

        valid_y = valid_y.reshape_out(d1 * d2, d3)
        yhat = yhat.reshape_out(d1 * d2, d3)
        valid_y = scaler.inverse_transform(valid_y)
        yhat = scaler.inverse_transform(yhat)

        valid_y = valid_y.reshape_out(d1, d2, d3)
        yhat = yhat.reshape_out(d1, d2, d3)

    dict_values = {}
    for a in range(len(config['y_colname'])):
        col_true, col_pred, col_mae, col_mse = [], [], [], []
        for i in range(config['n_output']):
            col_true.append(f'True_{i}')
            col_pred.append(f'Pred_{i}')
            col_mae.append(f'MAE_{i}')
            col_mse.append(f'MSE_{i}')
        col_value = col_true + col_pred + col_mae + col_mse
        df_value = pd.DataFrame(
            columns=col_value,
            index=range(valid_y.shape[0])
        )
        for jj in range(valid_y.shape[0]):
            list_true, list_pred, list_mae, list_mse = [], [], [], []
            for m in range(config['n_output']):
                true, pred = valid_y[jj][:, a: a + 1][m], yhat[jj][:, a: a + 1][m]
                list_true.append(true[0])
                list_pred.append(pred[0])
                list_mae.append(mean_absolute_error(true, pred))
                list_mse.append(mean_squared_error(true, pred))
            list_value = list_true + list_pred + list_mae + list_mse
            df_value.iloc[jj] = list_value

        if path_save != None:
            key = config['y_colname'][a]
            df_value.to_csv(
                f'{path_save}/{name_model}_{key}.csv',
                index=False
            )
        dict_values[config['y_colname'][a]] = df_value

    col_3d = ['MAE_3D', 'MSE_3D']
    col_2d = []
    for i in range(len(config['y_colname'])):
        col_2d.append(f'MAE_2D_V{i}')
        col_2d.append(f'MSE_2D_V{i}')

    col_3d_2d = col_3d + col_2d
    df_value_3d = pd.DataFrame(
        columns=col_3d_2d,
        index=range(valid_y.shape[0])
    )
    for ii in range(valid_y.shape[0]):
        list_mae_3d, list_mse_3d = [], []
        list_mae_3d.append(mean_absolute_error(valid_y[ii], yhat[ii]))
        list_mse_3d.append(mean_squared_error(valid_y[ii], yhat[ii]))

        list_mae_2d, list_mse_2d = [], []
        for a in range(len(config['y_colname'])):
            true, pred = valid_y[ii][:, a: a + 1], yhat[ii][:, a: a + 1]
            list_mae_2d.append(mean_absolute_error(true, pred))
            list_mse_2d.append(mean_squared_error(true, pred))
        list_value_3d_2d = list_mae_3d + list_mse_3d + list_mae_2d + list_mse_2d

        df_value_3d.iloc[ii] = list_value_3d_2d

    if path_save is not None:
        df_value.to_csv(
            f'{path_save}/{name_model}_3D.csv',
            index=False
        )
    return dict_values, df_value_3d


def plot_boxplot(dict_values, df_value_3d):
    metrics = ['MAE', 'MSE']
    for i in metrics:
        filter = [x for x in list(df_value_3d.columns) if re.match(i, x)]
        df_aux = df_value_3d[filter]
        fig, ax = plt.subplots(figsize=(10, 5))
        boxplot = sns.boxplot(data=df_aux.astype(float))
        # boxplot.axes.set_title("MAE en 3D y 2D", fontsize=16)
        boxplot.set_ylabel(i, fontsize=14)
        # boxplot.set_xlabel("Columnas", fontsize=14)
        plt.show()

    for k in metrics:
        i = 0
        fig, axes = plt.subplots(1, 4, figsize=(12, 3), constrained_layout=True)
        for x in list(dict_values.resolution()):
            filter = [x for x in dict_values[x].columns if re.match(k, x)]
            df_aux = dict_values[x][filter]
            fig = sns.boxplot(data=df_aux, ax=axes[i])
            fig.axes.set_title(x)
            fig.axes.set_ylabel(k)
            i += 1
        plt.show()



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm
import pandas as pd
import os
import pathlib
import openpy_fxts.models.imputation as its


def _convert_to_df(yhat, config_data, n_past):
    df_mask = pd.DataFrame(
        yhat.reshape(yhat.shape[0], yhat.shape[2]),
        index=config_data['dataset_orig'].iloc[n_past:].index,
        columns=config_data['y_colname']
    )
    df_miss = config_data['dataset_orig'][config_data['y_colname']]
    for i in df_miss.columns:
        df_miss[i][df_miss[i].isnull()] = df_mask[i].loc[df_miss[i].isnull()]
    df_miss.fillna(method="ffill")
    return df_miss


class run_all_models:

    def __init__(
            self,
            config_data_train: dict = None,
            config_data_test: dict = None,
            config_mdl: dict = None,
            config_sim: dict = None,
            config_arch: dict = None,
            source: str = None,
            type_mdl: str = None,
            name_mdl: str = None,
            train: int = None,
            test: int = None,
            folder: str = 'models_fxts',
            type_NaN: str = None,
            percentage: float = None,
            results: dict = None,
            view: bool = False
    ):
        self.config_data_train = config_data_train
        self.config_data_test = config_data_test
        self.config_mdl = config_mdl
        self.config_sim = config_sim
        self.config_arch = config_arch
        self.source = source
        self.type_mdl = type_mdl
        self.name_mdl = name_mdl
        self.train = train
        self.test = test
        self.folder = folder
        self.type_NaN = type_NaN
        self.percentage = percentage
        self.results = results
        self.view = view

    def exec_train(self, object_class):
        mdl = object_class(self.config_data_train, self.config_mdl, self.config_sim)
        print(mdl.name_mdl)
        path = pathlib.Path(self.source).joinpath(
            self.folder, self.type_NaN, f'pct_{self.percentage}', mdl.type_mdl, mdl.name_mdl
        )
        return mdl.training_mdl(path)

    def exec_test(self, object_class):
        mdl = object_class(self.config_data_test, self.config_mdl, self.config_sim)
        print(mdl.name_mdl)
        path = pathlib.Path(self.source).joinpath(
            self.folder, self.type_NaN, f'pct_{self.percentage}', mdl.type_mdl, mdl.name_mdl
        )
        new_model = mdl._build_mdl()
        list_files = [x for x in os.listdir(path) if 'weights' in x]
        '''
        aux_list = []
        characters = "weights-.hdf5"
        for file in list_files:
            for x in range(len(characters)):
                file = file.replace(characters[x], "")
            aux_list.append(file)
        '''
        print(f'Best Model => {list_files[-1]}')
        new_model.load_weights(str(path) + f'/{list_files[-1]}')
        yhat = mdl.prediction_mdl(new_model, path)
        self.results[mdl.name_mdl] = _convert_to_df(
            yhat,
            self.config_data_test,
            self.config_data_test['n_past']
        )
        if self.view:
            its.view.density_true_missing(
                self.config_data_test['dataset'][self.config_data_test['y_colname']],
                self.results,
                view=True,
                cumulative=False
            )
        return self.results

    def run_case_mdl(self, object_class):
        if self.train == 1 and self.test == 1:
            self.exec_train(object_class)
            return self.exec_test(object_class)
        if self.train == 1 and self.test == 0:
            return self.exec_train(object_class)
        if self.train == 0 and self.test == 1:
            return self.exec_test(object_class)

    def execute_case(self):
        if self.type_mdl == 'MLP':
            return self._mlp()
        if self.type_mdl == 'BiRNN':
            return self._birnn()
        if self.type_mdl == 'Conv1D':
            return self._conv1d()
        if self.type_mdl == 'Others':
            return self._others()
        if self.type_mdl == 'EncDec':
            return self._encdec()

        if self.type_mdl == 'LSTM':
            return self._lstm()
        if self.type_mdl == 'BiLSTM':
            return self._bilstm()
        if self.type_mdl == 'GRU':
            return self._gru()
        if self.type_mdl == 'Stacked':
            return self._stacked()
        if self.type_mdl == 'Seq2Seq':
            return self._seq2seq()
        else:
            return None

    def _mlp(self):
        if self.name_mdl == 'MLP_Dense':
            return self.run_case_mdl(its.MLP_Dense_class)

    def _birnn(self):
        # Simple models_fxts
        if self.name_mdl == 'BiRNN_Dense':
            return self.run_case_mdl(its.BiRNN_Dense_class)
        if self.name_mdl == 'Multi_BiRNN_Dense':
            return self.run_case_mdl(its.Multi_BiRNN_Dense_class)
        # Hybrids models_fxts
        if self.name_mdl == 'BiRNN_Conv1D':
            return self.run_case_mdl(its.BiRNN_Conv1D_class)
        if self.name_mdl == 'BiRNN_TimeDist_Dense':
            return self.run_case_mdl(its.BiRNN_TimeDist_Dense_class)
        if self.name_mdl == 'BiRNN_Bahdanau_Attention_Conv1D':
            return self.run_case_mdl(its.BiRNN_Bahdanau_Attention_Conv1D_class)
        if self.name_mdl == 'BiRNN_Luong_Attention_Conv1D':
            return self.run_case_mdl(its.BiRNN_Luong_Attention_Conv1D_class)
        if self.name_mdl == 'BiRNN_MultiHeadAttention_Conv1D':
            return self.run_case_mdl(its.BiRNN_MultiHeadAttention_Conv1D_class)
        if self.name_mdl == 'BiRNN_MDN':
            pass
            #return self.run_case_mdl(fxts.BiRNN_MDN)

    def _encdec(self):
        if self.name_mdl == 'EncDec_BiRNN':
            return self.run_case_mdl(its.EncDec_BiRNN_class)
        if self.name_mdl == 'EncDec_Conv1D_BiRNN':
            return self.run_case_mdl(its.EncDec_Conv1D_BiRNN_class)

    def _lstm(self):
        if self.name_mdl == 'LSTM2_Dense':
            return self.run_case_mdl(its.LSTM2_Dense_class)

    def _bilstm(self):
        if self.name_mdl == 'BiLSTM_Dense':
            return self.run_case_mdl(its.BiLSTM_Dense_class)
        if self.name_mdl == 'BiLSTM_Conv1D':
            return self.run_case_mdl(its.BiLSTM_Conv1D_class)
        if self.name_mdl == 'BiLSTM_Bahdanau_Attention_Conv1D':
            return self.run_case_mdl(its.BiLSTM_Bahdanau_Attention_Conv1D_class)
        if self.name_mdl == 'BiLSTM_MultiHeadAttention_Conv1D':
            return self.run_case_mdl(its.BiLSTM_MultiHeadAttention_Conv1D_class)
        if self.name_mdl == 'BiLSTM_Luong_Attention_Conv1D':
            return self.run_case_mdl(its.BiLSTM_Luong_Attention_Conv1D_class)
        if self.name_mdl == 'BiLSTM_MDN':
            return self.run_case_mdl(its.BiLSTM_MDN)  #Review

    def _conv1d(self):
        # Simple
        if self.name_mdl == 'Conv1D_Dense':
            return self.run_case_mdl(its.Conv1D_Dense_class)
        if self.name_mdl == 'Multi_Conv1D_Dense':
            return self.run_case_mdl(its.Multi_Conv1D_Dense_class)
        # Hybrids
        if self.name_mdl == 'Conv1D_BiRNN':
            return self.run_case_mdl(its.Conv1D_BiRNN_class)
        if self.name_mdl == 'TimeDist_Conv1D_BiRNN':
            return self.run_case_mdl(its.TimeDist_Conv1D_BiRNN_class)
        if self.name_mdl == 'Conv1D_BiRNN_Attention':
            return self.run_case_mdl(its.Conv1D_BiRNN_Attention_class)

    def _gru(self):
        if self.name_mdl == 'GRU2_Dense':
            return self.run_case_mdl(its.GRU_Dense_class)

    def _others(self):
        if self.name_mdl == 'TCN_BiRNN':
            return self.run_case_mdl(its.TCN_BiRNN_class)
        if self.name_mdl == 'Time2Vec_BiRNN':
            return self.run_case_mdl(its.Time2Vec_BiRNN_class)

    def _stacked(self):
        if self.name_mdl == 'LSTM_Stacked':
            return self.run_case_mdl(its.LSTM_Stacked)
        if self.name_mdl == 'BiLSTM_Stacked':
            return self.run_case_mdl(its.BiLSTM_Stacked)
        if self.name_mdl == 'Stacked_SciNet':
            return self.run_case_mdl(its.Stacked_SciNet)

    def _seq2seq(self):
        if self.name_mdl == 'Seq2Seq_LSTM':
            return self.run_case_mdl(its.Seq2Seq_LSTM_class)
        if self.name_mdl == 'Seq2Seq_LSTM2':
            return self.run_case_mdl(its.Seq2Seq_LSTM2_class)
        if self.name_mdl == 'Seq2Seq_LSTM_Batch_Drop':
            return self.run_case_mdl(its.Seq2Seq_LSTM_Batch_Drop_class)
        if self.name_mdl == 'Seq2Seq_BiLSTM':
            return self.run_case_mdl(its.Seq2Seq_BiLSTM_class)
        if self.name_mdl == 'Seq2Seq_BiLSTM2':
            return self.run_case_mdl(its.Seq2Seq_BiLSTM2_class)
        if self.name_mdl == 'Seq2Seq_Conv1D_BiLSTM':
            return self.run_case_mdl(its.Seq2Seq_Conv1D_BiLSTM_class)
        if self.name_mdl == 'Seq2Seq_Multi_Head_Conv1D_BiLSTM':
            return self.run_case_mdl(its.Seq2Seq_Multi_Head_Conv1D_BiLSTM_class)
        if self.name_mdl == 'Seq2Seq_LSTM_with_Luong_Attention':
            return self.run_case_mdl(its.Seq2Seq_LSTM_with_Luong_Attention_class)
        if self.name_mdl == 'Seq2Seq_BiLSTM_with_Attention':
            return self.run_case_mdl(its.Seq2Seq_BiLSTM_with_Attention_class)
        if self.name_mdl == 'Seq2Seq_Conv2D_LSTM':
            return self.run_case_mdl(its.Seq2Seq_Conv2D_LSTM_class)



# -*- coding: utf-8 -*-
# @Time    : 06/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.architectures.baseline_arch import base_arch
from openpy_fxts.architectures.fxts.fxts_mdls import fxts_mdls
from openpy_fxts.architectures.mdits.mdits_mdls import mdits_mdls


class get_architecture(base_arch):

    def __init__(self, name_mdl, type_mdl, n_past, n_future, n_inp_ft, n_out_ft, config_arch, app):
        super().__init__(name_mdl, type_mdl, n_past, n_future, n_inp_ft, n_out_ft, config_arch)
        self.app = app

    def select_model(self):

        if self.app == 'fxts':
            return fxts_mdls(
                self.name_mdl, self.type_mdl, self.n_past, self.n_future, self.n_inp_ft, self.n_out_ft, self.config_arch
            ).models()

        if self.app == 'mdits':
            return mdits_mdls(
                self.name_mdl, self.type_mdl, self.n_past, self.n_future, self.n_inp_ft, self.n_out_ft, self.config_arch
            ).models()





# -*- coding: utf-8 -*-
# @Time    : 04/09/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : evaluation.py
# @Software: PyCharm

class silhouette_coefficient:
    pass

class index_davis_buldin:
    pass




# -*- coding: utf-8 -*-
# @Time    : 27/04/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import tensorflow as tf
#from tf.keras.architectures import Model
#from tensorflow.keras.architectures import Model
#from tensorflow.keras.layers import Dense, Input, Flatten, Conv2D, MaxPool2D, LSTM, Bidirectional

# your config
config = {
    'learning_rate': 0.001,
    'lstm_neurons': 32,
    'lstm_activation': 'tanh',
    'dropout_rate': 0.08,
    'batch_size': 128,
    'dense_layers': [
        {'neurons': 32, 'activation': 'relu'},
        {'neurons': 32, 'activation': 'relu'},
    ]
}


# Subclassed API Model
class MySubClassed(tf.keras.Model):
    def __init__(self, output_size):
        super(MySubClassed, self).__init__()
        self.lstm = tf.keras.layers.LSTM(config['lstm_neurons'], activation=config['lstm_activation'])
        self.bn = tf.keras.layers.BatchNormalization()
        if 'dropout_rate' in config:
            self.dp1 = tf.keras.layers.Dropout(config['dropout_rate'])
            self.dp2 = tf.keras.layers.Dropout(config['dropout_rate'])
            self.dp3 = tf.keras.layers.Dropout(config['dropout_rate'])
        for layer in config['dense_layers']:
            self.dense1 = tf.keras.layers.Dense(layer['neurons'], activation=layer['activation'])
            self.bn1 = tf.keras.layers.BatchNormalization()
            self.dense2 = tf.keras.layers.Dense(layer['neurons'], activation=layer['activation'])
            self.bn2 = tf.keras.layers.BatchNormalization()
        self.out = tf.keras.layers.Dense(output_size, activation='sigmoid')

    def call(self, inputs, training=True, **kwargs):
        x = self.lstm(inputs)
        x = self.bn(x)

        if 'dropout_rate' in config:
            x = self.dp1(x)

        x = self.dense1(x)
        x = self.bn1(x)
        if 'dropout_rate' in config:
            x = self.dp2(x)

        x = self.dense2(x)
        x = self.bn2(x)
        if 'dropout_rate' in config:
            x = self.dp3(x)

        return self.out(x)

    # A convenient way to get model summary
    # and plot in subclassed api
    def build_graph(self, raw_shape):
        x = tf.keras.layers.Input(shape=(None, raw_shape), ragged=True)
        return tf.keras.Model(inputs=[x], outputs=self.call(x))

s = MySubClassed(output_size=1)
s.compile(
    loss='mse',
    metrics=['mse'],
    optimizer=tf.keras.optimizers.Adam(learning_rate=0.001)
)

raw_input = (16, 16, 16)
y = s(tf.ones(shape=(raw_input)))

print("weights:", len(s.weights))
print("trainable weights:", len(s.trainable_weights))

s.build_graph(16).summary()
tf.keras.utils.plot_model(
    s.build_graph(16),
    show_shapes=True,
    show_dtype=True,
    show_layer_names=True,
    rankdir="TB",
)


class LeNet5(tf.keras.Model):
    def __init__(self):
        super(LeNet5, self).__init__()
        #creating layers in initializer
        self.conv1 = tf.keras.layers.Conv2D(filters=6, kernel_size=(5, 5), padding="same", activation="relu")
        self.max_pool2x2 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))
        self.conv2 = tf.keras.layers.Conv2D(filters=16, kernel_size=(5, 5), padding="same", activation="relu")
        self.conv3 = tf.keras.layers.Conv2D(filters=120, kernel_size=(5, 5), padding="same", activation="relu")
        self.flatten = tf.keras.layers.Flatten()
        self.fc2 = tf.keras.layers.Dense(units=84, activation="relu")
        self.fc3 = tf.keras.layers.Dense(units=10, activation="softmax")

    def call(self, input_tensor):
        #don't add layers here, need to create the layers in initializer, otherwise you will get the tf.Variable can only be created once error
        x = self.conv1(input_tensor)
        x = self.max_pool2x2(x)
        x = self.conv2(x)
        x = self.max_pool2x2(x)
        x = self.conv3(x)
        x = self.flatten(x)
        x = self.fc2(x)
        x = self.fc3(x)
        return x

input_layer = tf.keras.layers.Input(shape=(32,32,3,))
x = LeNet5()(input_layer)

model = tf.keras.Model(inputs=input_layer, outputs=x)
print(model.summary(expand_nested=True))

class LSTM_model(tf.keras.Model):

    def __init__(self, units: int = 50, activation: str ="tanh"):
        super(LSTM_model, self).__init__()
        self.lstm1 = tf.keras.layers.LSTM(
            units=units,
            activation=activation
        )

    def call(self, inputs):
        x = self.lstm1(inputs)
        return tf.keras.Model(inputs=[x], outputs=self.call(x))


class Conv1D_Dense_class(tf.keras.Model):

    def __init__(self,):
        super(Conv1D_Dense, self).__init__()
        self.conv1d = tf.keras.layers.Conv1D(filters=64, kernel_size=2, activation='relu')
        self.maxpool1d = tf.keras.layers.MaxPooling1D(pool_size=2)
        self.flatten = tf.keras.layers.Flatten()
        self.dense1b = tf.keras.layers.Dense(units=50, activation="relu")



def Conv1D_Dense(n_steps_in, n_features, n_output):

    model = tf.keras.Sequential()
    model.add(tf.keras.layers.Conv1D(filters=64, kernel_size=2, activation='relu', input_shape=(n_steps_in, n_features)))
    model.add(tf.keras.layers.MaxPooling1D(pool_size=2))
    model.add(tf.keras.layers.Flatten())
    model.add(tf.keras.layers.Dense(50, activation='relu'))
    model.add(tf.keras.layers.Dense(n_output))
    model.add(tf.keras.layers.Reshape())
    return model

# multivariate output multi-step 1d cnn example
from numpy import array
from numpy import hstack
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import Flatten
from keras.layers.convolutional import Conv1D
from keras.layers.convolutional import MaxPooling1D

def split_sequences(sequences, n_steps_in, n_steps_out):
    X, y = list(), list()
    for i in range(len(sequences)):
        # find the end of this pattern
        end_ix = i + n_steps_in
        out_end_ix = end_ix + n_steps_out
        # check if we are beyond the dataset
        if out_end_ix > len(sequences):
            break
        # gather input and output parts of the pattern
        seq_x, seq_y = sequences[i:end_ix, :], sequences[end_ix:out_end_ix, :]
        X.append(seq_x)
        y.append(seq_y)
    return array(X), array(y)

# define input sequence
in_seq1 = array([10, 20, 30, 40, 50, 60, 70, 80, 90])
in_seq2 = array([15, 25, 35, 45, 55, 65, 75, 85, 95])
out_seq = array([in_seq1[i]+in_seq2[i] for i in range(len(in_seq1))])
# convert to [rows, columns] structure
in_seq1 = in_seq1.reshape((len(in_seq1), 1))
in_seq2 = in_seq2.reshape((len(in_seq2), 1))
out_seq = out_seq.reshape((len(out_seq), 1))
# horizontally stack columns
dataset = hstack((in_seq1, in_seq2, out_seq))
# choose a number of time steps
n_steps_in, n_steps_out = 3, 2
# convert into input/output
X, y = split_sequences(dataset, n_steps_in, n_steps_out)
print(X.shape, y.shape)

# flatten output
n_output = y.shape[1] * y.shape[2]
y = y.reshape((y.shape[0], n_output))
# the dataset knows the number of features, e.g. 2
n_features = X.shape[2]
# define model


model = Conv1D_Dense(n_steps_in, n_features, n_output)

model.compile(optimizer='adam', loss='mse')
# fit model
model.fit(X, y, epochs=7000, verbose=0)
# demonstrate prediction
x_input = array([[60, 65, 125], [70, 75, 145], [80, 85, 165]])
print(f'x_input:\n{x_input}')
x_input = x_input.reshape((1, n_steps_in, n_features))
yhat = model.predict(x_input, verbose=0)
print(f'yhat:\n{yhat}')


from openpy_fxts.base_tf import tkm, tkl


class _decoder_lstm(tkm.Model):
    def __init__(self, n_out_ft=None):
        super(_decoder_lstm, self).__init__()
        self.lstm = tkl.LSTM(256, return_sequences=True)
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        decoder = self.lstm(inputs[0], initial_state=inputs[1])
        decoder = self.time_dense(decoder)
        return decoder


class _decoder_lstm2(tkm.Model):
    def __init__(self, n_out_ft=None):
        super(_decoder_lstm2, self).__init__()
        self.decoder_l1 = tkl.LSTM(256, return_sequences=True)
        self.decoder_l2 = tkl.LSTM(256, return_sequences=True)
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        decoder_l1 = self.decoder_l1(inputs[0], initial_state=inputs[1])
        decoder_l2 = self.decoder_l2(decoder_l1, initial_state=inputs[2])
        decoder = self.time_dense(decoder_l2)
        return decoder


class _decoder_lstm2_bach_drop(tkm.Model):

    def __init__(self, n_out_ft=None):
        super(_decoder_lstm2_bach_drop, self).__init__()
        self.batch = tkl.BatchNormalization()
        self.lstm1 = tkl.LSTM(256, return_sequences=True)
        self.drop = tkl.Dropout(0.2)
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        decoder_inputs = self.batch(inputs)
        decoder_out = self.lstm1(decoder_inputs)
        decoder_out = self.drop(decoder_out)
        decoder_out = self.batch(decoder_out)
        decoder_out = self.time_dense(decoder_out)
        return decoder_out


class _decoder_bilstm(tkm.Model):
    def __init__(self, n_out_ft=None):
        super(_decoder_bilstm, self).__init__()
        self.bilstm = tkl.Bidirectional(tkl.LSTM(256, return_sequences=True))
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        decoder = self.bilstm(inputs[0], initial_state=inputs[1])
        decoder = self.time_dense(decoder)
        return decoder


class _decoder_bilstm2(tkm.Model):
    def __init__(self, n_out_ft=None):
        super(_decoder_bilstm2, self).__init__()
        self.decoder_l1 = tkl.Bidirectional(tkl.LSTM(256, return_sequences=True))
        self.decoder_l2 = tkl.Bidirectional(tkl.LSTM(256, return_sequences=True))
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        decoder_l1 = self.decoder_l1(inputs[0], initial_state=inputs[1])
        decoder_l2 = self.decoder_l2(decoder_l1, initial_state=inputs[2])
        decoder = self.time_dense(decoder_l2)
        return decoder


class _decoder_bilstm_time_dist(tkm.Model):
    def __init__(self, n_out_ft=None):
        super(_decoder_bilstm_time_dist, self).__init__()
        self.bilstm = tkl.Bidirectional(tkl.LSTM(256, activation='relu', return_sequences=True))
        self.time_dense1 = tkl.TimeDistributed(tkl.Dense(256, activation='relu'))
        self.time_dense2 = tkl.TimeDistributed(tkl.Dense(n_out_ft, activation='relu'))

    def call(self, inputs, training=True, **kwargs):
        x = self.bilstm(inputs)
        x = self.time_dense1(x)
        x = self.time_dense2(x)
        return x

class _decoder_bilstm(tkm.Model):
    def __init__(self, n_out_ft=None):
        super(_decoder_bilstm, self).__init__()
        self.bilstm = tkl.Bidirectional(tkl.LSTM(256, return_sequences=True))
        self.time_dense = tkl.TimeDistributed(tkl.Dense(n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        decoder = self.bilstm(inputs[0], initial_state=inputs[1])
        decoder = self.time_dense(decoder)
        return decoder




# -*- coding: utf-8 -*-
# @Time    : 25/03/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import numpy as np
import pandas as pd
import sklearn.compose
import sklearn.impute
import sklearn.preprocessing

from sklearn.linear_model import BayesianRidge
from fancyimpute import IterativeSVD
from openpy_fxts.models.imputation.base_lib import init_models


from openpy_fxts.models.imputation.utilities.Utils import Preprocessing, Metrics, Utils

prepro = Preprocessing()
metrics = Metrics()


class iterative(init_models):

    def __init__(self, df_miss: pd.DataFrame = None, df_true: pd.DataFrame = None):
        super().__init__(df_miss, df_true)
        self.df_miss = df_miss
        self.df_true = df_true

    def MICE(
            self,
            estimator=BayesianRidge(),
            missing_values=np.nan,
            sample_posterior=False,
            max_iter=10,
            tol=1e-3,
            n_nearest_features=None,
            initial_strategy="mean",
            imputation_order="ascending",
            skip_complete=False,
            min_value=-np.inf,
            max_value=np.inf,
            verbose=0,
            random_state=None,
            add_indicator=False,
            keep_empty_features=False
    ):
        list_aux = Utils(self.df_miss).list_column_missing()
        df_miss, transf = prepro.LabelEncoder(self.df_miss)

        model_mice = sklearn.impute.IterativeImputer(
            estimator=estimator,
            missing_values=missing_values,
            sample_posterior=sample_posterior,
            max_iter=max_iter,
            tol=tol,
            n_nearest_features=n_nearest_features,
            initial_strategy=initial_strategy,
            imputation_order=imputation_order,
            skip_complete=skip_complete,
            min_value=min_value,
            max_value=max_value,
            verbose=verbose,
            random_state=random_state,
            add_indicator=add_indicator,
            keep_empty_features=keep_empty_features
        )
        df_mice = df_miss.copy(deep=True)
        df_mice.iloc[:, :] = model_mice.fit_transform(df_mice).round()
        if self.df_true is None:
            return df_mice
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(self.df_true[list_aux], df_mice[list_aux], dict_aux, 'mice_forest')
            return df_mice, dict_aux['mice_forest']

    def iter_SVD(
            self,
            rank=10,
            convergence_threshold=0.00001,
            max_iters=200,
            gradual_rank_increase=True,
            svd_algorithm="arpack",
            init_fill_method="zero",
            min_value=None,
            max_value=None,
            verbose=True
    ):
        list_aux = Utils(self.df_miss).list_column_missing()
        self.df_miss, transf = prepro.LabelEncoder(self.df_miss)
        model_svd = IterativeSVD(
            rank,
            convergence_threshold,
            max_iters,
            gradual_rank_increase,
            svd_algorithm,
            init_fill_method,
            min_value,
            max_value,
            verbose
        )

        X_miss = self.df_miss.to_numpy()
        svd_imp = model_svd.fit_transform(X_miss)
        svd_imp = pd.DataFrame(svd_imp, columns=self.df_miss.columns)
        svd_imp['datetime'] = self.df_miss.index
        svd_imp.set_index('datetime', inplace=True)

        if self.df_true is None:
            return svd_imp
        else:
            dict_aux = dict()
            dict_aux = metrics.add_dict_metrics(self.df_true[list_aux], svd_imp[list_aux], dict_aux, 'IterativeSVD')
            return svd_imp, dict_aux['IterativeSVD']
